"""
    :codeauthor: :email:`Daniel Wallace <dwallace@saltstack.com`
"""

import os
import re
import shutil
import tempfile

import pytest
import salt.client.ssh.client
import salt.config
import salt.roster
import salt.utils.files
import salt.utils.path
import salt.utils.thin
import salt.utils.yaml
from salt.client import ssh
from tests.support.case import ShellCase
from tests.support.mock import MagicMock, call, patch
from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import TestCase, skipIf


@skipIf(not salt.utils.path.which("ssh"), "No ssh binary found in path")
class SSHPasswordTests(ShellCase):
    @pytest.mark.slow_test
    def test_password_failure(self):
        """
        Check password failures when trying to deploy keys
        """
        opts = salt.config.client_config(self.get_config_file_path("master"))
        opts["list_hosts"] = False
        opts["argv"] = ["test.ping"]
        opts["selected_target_option"] = "glob"
        opts["tgt"] = "localhost"
        opts["arg"] = []
        roster = os.path.join(RUNTIME_VARS.TMP_CONF_DIR, "roster")
        handle_ssh_ret = [
            {
                "localhost": {
                    "retcode": 255,
                    "stderr": "Permission denied (publickey).\r\n",
                    "stdout": "",
                }
            },
        ]
        expected = {"localhost": "Permission denied (publickey)"}
        display_output = MagicMock()
        with patch(
            "salt.roster.get_roster_file", MagicMock(return_value=roster)
        ), patch(
            "salt.client.ssh.SSH.handle_ssh", MagicMock(return_value=handle_ssh_ret)
        ), patch(
            "salt.client.ssh.SSH.key_deploy", MagicMock(return_value=expected)
        ), patch(
            "salt.output.display_output", display_output
        ):
            client = ssh.SSH(opts)
            ret = next(client.run_iter())
            with self.assertRaises(SystemExit):
                client.run()
        display_output.assert_called_once_with(expected, "nested", opts)
        self.assertIs(ret, handle_ssh_ret[0])


@skipIf(not salt.utils.path.which("ssh"), "No ssh binary found in path")
class SSHReturnEventTests(ShellCase):
    def test_not_missing_fun_calling_wfuncs(self):
        opts = salt.config.client_config(self.get_config_file_path("master"))
        opts["list_hosts"] = False
        opts["argv"] = ["state.show_highstate"]
        opts["selected_target_option"] = "glob"
        opts["tgt"] = "localhost"
        opts["arg"] = []
        roster = os.path.join(RUNTIME_VARS.TMP_CONF_DIR, "roster")
        handle_ssh_ret = [
            {"localhost": {}},
        ]

        expected = {"localhost": {}}
        display_output = MagicMock()
        with patch(
            "salt.roster.get_roster_file", MagicMock(return_value=roster)
        ), patch(
            "salt.client.ssh.SSH.handle_ssh", MagicMock(return_value=handle_ssh_ret)
        ), patch(
            "salt.client.ssh.SSH.key_deploy", MagicMock(return_value=expected)
        ), patch(
            "salt.output.display_output", display_output
        ):
            client = ssh.SSH(opts)
            client.event = MagicMock()
            ret = next(client.run_iter())
            assert "localhost" in ret
            assert "fun" in ret["localhost"]
            client.run()
        display_output.assert_called_once_with(expected, "nested", opts)
        self.assertIs(ret, handle_ssh_ret[0])
        assert len(client.event.fire_event.call_args_list) == 2
        assert "fun" in client.event.fire_event.call_args_list[0][0][0]
        assert "fun" in client.event.fire_event.call_args_list[1][0][0]


class SSHRosterDefaults(TestCase):
    def setUp(self):
        self.roster = """
            localhost:
              host: 127.0.0.1
              port: 2827
            self:
              host: 0.0.0.0
              port: 42
            """

    def test_roster_defaults_flat(self):
        """
        Test Roster Defaults on the flat roster
        """
        tempdir = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)
        expected = {
            "self": {"host": "0.0.0.0", "user": "daniel", "port": 42},
            "localhost": {"host": "127.0.0.1", "user": "daniel", "port": 2827},
        }
        try:
            root_dir = os.path.join(tempdir, "foo", "bar")
            os.makedirs(root_dir)
            fpath = os.path.join(root_dir, "config")
            with salt.utils.files.fopen(fpath, "w") as fp_:
                fp_.write(
                    """
                    roster_defaults:
                      user: daniel
                    """
                )
            opts = salt.config.master_config(fpath)
            with patch(
                "salt.roster.get_roster_file", MagicMock(return_value=self.roster)
            ):
                with patch(
                    "salt.template.compile_template",
                    MagicMock(return_value=salt.utils.yaml.safe_load(self.roster)),
                ):
                    roster = salt.roster.Roster(opts=opts)
                    self.assertEqual(roster.targets("*", "glob"), expected)
        finally:
            if os.path.isdir(tempdir):
                shutil.rmtree(tempdir)


class SSHSingleTests(TestCase):
    def setUp(self):
        self.tmp_cachedir = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)
        self.argv = [
            "ssh.set_auth_key",
            "root",
            "hobn+amNAXSBTiOXEqlBjGB...rsa root@master",
        ]
        self.opts = {
            "argv": self.argv,
            "__role": "master",
            "cachedir": self.tmp_cachedir,
            "extension_modules": os.path.join(self.tmp_cachedir, "extmods"),
        }
        self.target = {
            "passwd": "abc123",
            "ssh_options": None,
            "sudo": False,
            "identities_only": False,
            "host": "login1",
            "user": "root",
            "timeout": 65,
            "remote_port_forwards": None,
            "sudo_user": "",
            "port": "22",
            "priv": "/etc/salt/pki/master/ssh/salt-ssh.rsa",
        }

    def test_single_opts(self):
        """Sanity check for ssh.Single options"""

        single = ssh.Single(
            self.opts,
            self.opts["argv"],
            "localhost",
            mods={},
            fsclient=None,
            thin=salt.utils.thin.thin_path(self.opts["cachedir"]),
            mine=False,
            **self.target
        )

        self.assertEqual(single.shell._ssh_opts(), "")
        self.assertEqual(
            single.shell._cmd_str("date +%s"),
            "ssh login1 "
            "-o KbdInteractiveAuthentication=no -o "
            "PasswordAuthentication=yes -o ConnectTimeout=65 -o Port=22 "
            "-o IdentityFile=/etc/salt/pki/master/ssh/salt-ssh.rsa "
            "-o User=root  date +%s",
        )

    def test_run_with_pre_flight(self):
        """
        test Single.run() when ssh_pre_flight is set
        and script successfully runs
        """
        target = self.target.copy()
        target["ssh_pre_flight"] = os.path.join(RUNTIME_VARS.TMP, "script.sh")
        single = ssh.Single(
            self.opts,
            self.opts["argv"],
            "localhost",
            mods={},
            fsclient=None,
            thin=salt.utils.thin.thin_path(self.opts["cachedir"]),
            mine=False,
            **target
        )

        cmd_ret = ("Success", "", 0)
        mock_flight = MagicMock(return_value=cmd_ret)
        mock_cmd = MagicMock(return_value=cmd_ret)
        patch_flight = patch("salt.client.ssh.Single.run_ssh_pre_flight", mock_flight)
        patch_cmd = patch("salt.client.ssh.Single.cmd_block", mock_cmd)
        patch_exec_cmd = patch(
            "salt.client.ssh.shell.Shell.exec_cmd", return_value=("", "", 1)
        )
        patch_os = patch("os.path.exists", side_effect=[True])

        with patch_os, patch_flight, patch_cmd, patch_exec_cmd:
            ret = single.run()
            mock_cmd.assert_called()
            mock_flight.assert_called()
            assert ret == cmd_ret

    def test_run_with_pre_flight_stderr(self):
        """
        test Single.run() when ssh_pre_flight is set
        and script errors when run
        """
        target = self.target.copy()
        target["ssh_pre_flight"] = os.path.join(RUNTIME_VARS.TMP, "script.sh")
        single = ssh.Single(
            self.opts,
            self.opts["argv"],
            "localhost",
            mods={},
            fsclient=None,
            thin=salt.utils.thin.thin_path(self.opts["cachedir"]),
            mine=False,
            **target
        )

        cmd_ret = ("", "Error running script", 1)
        mock_flight = MagicMock(return_value=cmd_ret)
        mock_cmd = MagicMock(return_value=cmd_ret)
        patch_flight = patch("salt.client.ssh.Single.run_ssh_pre_flight", mock_flight)
        patch_cmd = patch("salt.client.ssh.Single.cmd_block", mock_cmd)
        patch_exec_cmd = patch(
            "salt.client.ssh.shell.Shell.exec_cmd", return_value=("", "", 1)
        )
        patch_os = patch("os.path.exists", side_effect=[True])

        with patch_os, patch_flight, patch_cmd, patch_exec_cmd:
            ret = single.run()
            mock_cmd.assert_not_called()
            mock_flight.assert_called()
            assert ret == cmd_ret

    def test_run_with_pre_flight_script_doesnot_exist(self):
        """
        test Single.run() when ssh_pre_flight is set
        and the script does not exist
        """
        target = self.target.copy()
        target["ssh_pre_flight"] = os.path.join(RUNTIME_VARS.TMP, "script.sh")
        single = ssh.Single(
            self.opts,
            self.opts["argv"],
            "localhost",
            mods={},
            fsclient=None,
            thin=salt.utils.thin.thin_path(self.opts["cachedir"]),
            mine=False,
            **target
        )

        cmd_ret = ("Success", "", 0)
        mock_flight = MagicMock(return_value=cmd_ret)
        mock_cmd = MagicMock(return_value=cmd_ret)
        patch_flight = patch("salt.client.ssh.Single.run_ssh_pre_flight", mock_flight)
        patch_cmd = patch("salt.client.ssh.Single.cmd_block", mock_cmd)
        patch_exec_cmd = patch(
            "salt.client.ssh.shell.Shell.exec_cmd", return_value=("", "", 1)
        )
        patch_os = patch("os.path.exists", side_effect=[False])

        with patch_os, patch_flight, patch_cmd, patch_exec_cmd:
            ret = single.run()
            mock_cmd.assert_called()
            mock_flight.assert_not_called()
            assert ret == cmd_ret

    def test_run_with_pre_flight_thin_dir_exists(self):
        """
        test Single.run() when ssh_pre_flight is set
        and thin_dir already exists
        """
        target = self.target.copy()
        target["ssh_pre_flight"] = os.path.join(RUNTIME_VARS.TMP, "script.sh")
        single = ssh.Single(
            self.opts,
            self.opts["argv"],
            "localhost",
            mods={},
            fsclient=None,
            thin=salt.utils.thin.thin_path(self.opts["cachedir"]),
            mine=False,
            **target
        )

        cmd_ret = ("", "", 0)
        mock_flight = MagicMock(return_value=cmd_ret)
        mock_cmd = MagicMock(return_value=cmd_ret)
        patch_flight = patch("salt.client.ssh.Single.run_ssh_pre_flight", mock_flight)
        patch_cmd = patch("salt.client.ssh.shell.Shell.exec_cmd", mock_cmd)
        patch_cmd_block = patch("salt.client.ssh.Single.cmd_block", mock_cmd)
        patch_os = patch("os.path.exists", return_value=True)

        with patch_os, patch_flight, patch_cmd, patch_cmd_block:
            ret = single.run()
            mock_cmd.assert_called()
            mock_flight.assert_not_called()
            assert ret == cmd_ret

    def test_execute_script(self):
        """
        test Single.execute_script()
        """
        single = ssh.Single(
            self.opts,
            self.opts["argv"],
            "localhost",
            mods={},
            fsclient=None,
            thin=salt.utils.thin.thin_path(self.opts["cachedir"]),
            mine=False,
            winrm=False,
            **self.target
        )

        exp_ret = ("Success", "", 0)
        mock_cmd = MagicMock(return_value=exp_ret)
        patch_cmd = patch("salt.client.ssh.shell.Shell.exec_cmd", mock_cmd)
        script = os.path.join(RUNTIME_VARS.TMP, "script.sh")

        with patch_cmd:
            ret = single.execute_script(script=script)
            assert ret == exp_ret
            assert mock_cmd.call_count == 2
            assert [
                call("/bin/sh '{}'".format(script)),
                call("rm '{}'".format(script)),
            ] == mock_cmd.call_args_list

    def test_shim_cmd(self):
        """
        test Single.shim_cmd()
        """
        single = ssh.Single(
            self.opts,
            self.opts["argv"],
            "localhost",
            mods={},
            fsclient=None,
            thin=salt.utils.thin.thin_path(self.opts["cachedir"]),
            mine=False,
            winrm=False,
            tty=True,
            **self.target
        )

        exp_ret = ("Success", "", 0)
        mock_cmd = MagicMock(return_value=exp_ret)
        patch_cmd = patch("salt.client.ssh.shell.Shell.exec_cmd", mock_cmd)
        patch_send = patch("salt.client.ssh.shell.Shell.send", return_value=("", "", 0))
        patch_rand = patch("os.urandom", return_value=b"5\xd9l\xca\xc2\xff")

        with patch_cmd, patch_rand, patch_send:
            ret = single.shim_cmd(cmd_str="echo test")
            assert ret == exp_ret
            assert [
                call("/bin/sh '.35d96ccac2ff.py'"),
                call("rm '.35d96ccac2ff.py'"),
            ] == mock_cmd.call_args_list

    def test_run_ssh_pre_flight(self):
        """
        test Single.run_ssh_pre_flight
        """
        target = self.target.copy()
        target["ssh_pre_flight"] = os.path.join(RUNTIME_VARS.TMP, "script.sh")
        single = ssh.Single(
            self.opts,
            self.opts["argv"],
            "localhost",
            mods={},
            fsclient=None,
            thin=salt.utils.thin.thin_path(self.opts["cachedir"]),
            mine=False,
            winrm=False,
            tty=True,
            **target
        )

        exp_ret = ("Success", "", 0)
        mock_cmd = MagicMock(return_value=exp_ret)
        patch_cmd = patch("salt.client.ssh.shell.Shell.exec_cmd", mock_cmd)
        patch_send = patch("salt.client.ssh.shell.Shell.send", return_value=exp_ret)
        exp_tmp = os.path.join(
            tempfile.gettempdir(), os.path.basename(target["ssh_pre_flight"])
        )

        with patch_cmd, patch_send:
            ret = single.run_ssh_pre_flight()
            assert ret == exp_ret
            assert [
                call("/bin/sh '{}'".format(exp_tmp)),
                call("rm '{}'".format(exp_tmp)),
            ] == mock_cmd.call_args_list

    @skipIf(salt.utils.platform.is_windows(), "SSH_PY_SHIM not set on windows")
    def test_cmd_run_set_path(self):
        """
        test when set_path is set
        """
        target = self.target
        target["set_path"] = "$PATH:/tmp/path/"
        single = ssh.Single(
            self.opts,
            self.opts["argv"],
            "localhost",
            mods={},
            fsclient=None,
            thin=salt.utils.thin.thin_path(self.opts["cachedir"]),
            mine=False,
            **self.target
        )

        ret = single._cmd_str()
        assert re.search("\\" + target["set_path"], ret)

    @skipIf(salt.utils.platform.is_windows(), "SSH_PY_SHIM not set on windows")
    def test_cmd_run_not_set_path(self):
        """
        test when set_path is not set
        """
        target = self.target
        single = ssh.Single(
            self.opts,
            self.opts["argv"],
            "localhost",
            mods={},
            fsclient=None,
            thin=salt.utils.thin.thin_path(self.opts["cachedir"]),
            mine=False,
            **self.target
        )

        ret = single._cmd_str()
        assert re.search('SET_PATH=""', ret)


@skipIf(not salt.utils.path.which("ssh"), "No ssh binary found in path")
class SSHTests(ShellCase):
    def setUp(self):
        self.roster = """
            localhost:
              host: 127.0.0.1
              port: 2827
            """
        self.opts = salt.config.client_config(self.get_config_file_path("master"))
        self.opts["selected_target_option"] = "glob"

    def test_expand_target_ip_address(self):
        """
        test expand_target when target is root@<ip address>
        """
        host = "127.0.0.1"
        user = "test-user@"
        opts = self.opts
        opts["tgt"] = user + host

        with patch(
            "salt.utils.network.is_reachable_host", MagicMock(return_value=False)
        ):
            client = ssh.SSH(opts)
        assert opts["tgt"] == user + host
        with patch(
            "salt.roster.get_roster_file", MagicMock(return_value="/etc/salt/roster")
        ), patch(
            "salt.client.ssh.compile_template",
            MagicMock(return_value=salt.utils.yaml.safe_load(self.roster)),
        ):
            client._expand_target()
        assert opts["tgt"] == host

    def test_expand_target_dns(self):
        """
        test expand_target when target is root@<dns>
        """
        host = "localhost"
        user = "test-user@"
        opts = self.opts
        opts["tgt"] = user + host

        with patch(
            "salt.utils.network.is_reachable_host", MagicMock(return_value=False)
        ):
            client = ssh.SSH(opts)
        assert opts["tgt"] == user + host
        with patch(
            "salt.roster.get_roster_file", MagicMock(return_value="/etc/salt/roster")
        ), patch(
            "salt.client.ssh.compile_template",
            MagicMock(return_value=salt.utils.yaml.safe_load(self.roster)),
        ):
            client._expand_target()
        assert opts["tgt"] == host

    def test_expand_target_no_user(self):
        """
        test expand_target when no user defined
        """
        host = "127.0.0.1"
        opts = self.opts
        opts["tgt"] = host

        with patch(
            "salt.utils.network.is_reachable_host", MagicMock(return_value=False)
        ):
            client = ssh.SSH(opts)
        assert opts["tgt"] == host

        with patch(
            "salt.roster.get_roster_file", MagicMock(return_value="/etc/salt/roster")
        ), patch(
            "salt.client.ssh.compile_template",
            MagicMock(return_value=salt.utils.yaml.safe_load(self.roster)),
        ):
            client._expand_target()
        assert opts["tgt"] == host

    def test_update_targets_ip_address(self):
        """
        test update_targets when host is ip address
        """
        host = "127.0.0.1"
        user = "test-user@"
        opts = self.opts
        opts["tgt"] = user + host

        with patch(
            "salt.utils.network.is_reachable_host", MagicMock(return_value=False)
        ):
            client = ssh.SSH(opts)
        assert opts["tgt"] == user + host
        client._update_targets()
        assert opts["tgt"] == host
        assert client.targets[host]["user"] == user.split("@")[0]

    def test_update_targets_dns(self):
        """
        test update_targets when host is dns
        """
        host = "localhost"
        user = "test-user@"
        opts = self.opts
        opts["tgt"] = user + host

        with patch(
            "salt.utils.network.is_reachable_host", MagicMock(return_value=False)
        ):
            client = ssh.SSH(opts)
        assert opts["tgt"] == user + host
        client._update_targets()
        assert opts["tgt"] == host
        assert client.targets[host]["user"] == user.split("@")[0]

    def test_update_targets_no_user(self):
        """
        test update_targets when no user defined
        """
        host = "127.0.0.1"
        opts = self.opts
        opts["tgt"] = host

        with patch(
            "salt.utils.network.is_reachable_host", MagicMock(return_value=False)
        ):
            client = ssh.SSH(opts)
        assert opts["tgt"] == host
        client._update_targets()
        assert opts["tgt"] == host

    def test_update_expand_target_dns(self):
        """
        test update_targets and expand_target when host is dns
        """
        host = "localhost"
        user = "test-user@"
        opts = self.opts
        opts["tgt"] = user + host

        with patch(
            "salt.utils.network.is_reachable_host", MagicMock(return_value=False)
        ):
            client = ssh.SSH(opts)
        assert opts["tgt"] == user + host
        with patch(
            "salt.roster.get_roster_file", MagicMock(return_value="/etc/salt/roster")
        ), patch(
            "salt.client.ssh.compile_template",
            MagicMock(return_value=salt.utils.yaml.safe_load(self.roster)),
        ):
            client._expand_target()
        client._update_targets()
        assert opts["tgt"] == host
        assert client.targets[host]["user"] == user.split("@")[0]

    def test_parse_tgt(self):
        """
        test parse_tgt when user and host set on
        the ssh cli tgt
        """
        host = "localhost"
        user = "test-user@"
        opts = self.opts
        opts["tgt"] = user + host

        with patch(
            "salt.utils.network.is_reachable_host", MagicMock(return_value=False)
        ):
            assert not self.opts.get("ssh_cli_tgt")
            client = ssh.SSH(opts)
            assert client.parse_tgt["hostname"] == host
            assert client.parse_tgt["user"] == user.split("@")[0]
            assert self.opts.get("ssh_cli_tgt") == user + host

    def test_parse_tgt_no_user(self):
        """
        test parse_tgt when only the host set on
        the ssh cli tgt
        """
        host = "localhost"
        opts = self.opts
        opts["ssh_user"] = "ssh-usr"
        opts["tgt"] = host

        with patch(
            "salt.utils.network.is_reachable_host", MagicMock(return_value=False)
        ):
            assert not self.opts.get("ssh_cli_tgt")
            client = ssh.SSH(opts)
            assert client.parse_tgt["hostname"] == host
            assert client.parse_tgt["user"] == opts["ssh_user"]
            assert self.opts.get("ssh_cli_tgt") == host

    def test_extra_filerefs(self):
        """
        test "extra_filerefs" are not excluded from kwargs
        when preparing the SSH opts
        """
        opts = {
            "eauth": "auto",
            "username": "test",
            "password": "test",
            "client": "ssh",
            "tgt": "localhost",
            "fun": "test.ping",
            "ssh_port": 22,
            "extra_filerefs": "salt://foobar",
        }
        roster = os.path.join(RUNTIME_VARS.TMP_CONF_DIR, "roster")
        client = salt.client.ssh.client.SSHClient(
            mopts=self.opts, disable_custom_roster=True
        )
        with patch("salt.roster.get_roster_file", MagicMock(return_value=roster)):
            ssh_obj = client._prep_ssh(**opts)
            assert ssh_obj.opts.get("extra_filerefs", None) == "salt://foobar"



"""
Work with cron

.. note::
    Salt does not escape cron metacharacters automatically. You should
    backslash-escape percent characters and any other metacharacters that might
    be interpreted incorrectly by the shell.
"""

import logging
import os
import random

import salt.utils.data
import salt.utils.files
import salt.utils.functools
import salt.utils.path
import salt.utils.stringutils

TAG = "# Lines below here are managed by Salt, do not edit\n"
SALT_CRON_IDENTIFIER = "SALT_CRON_IDENTIFIER"
SALT_CRON_NO_IDENTIFIER = "NO ID SET"

log = logging.getLogger(__name__)


def __virtual__():
    if salt.utils.path.which("crontab"):
        return True
    else:
        return (False, "Cannot load cron module: crontab command not found")


def _ensure_string(val):
    # Account for cases where the identifier is not a string
    # which would cause to_unicode to fail.
    if not isinstance(val, str):
        val = str(val)
    try:
        return salt.utils.stringutils.to_unicode(val)
    except TypeError:
        return ""


def _cron_id(cron):
    """SAFETYBELT, Only set if we really have an identifier"""
    cid = None
    if cron["identifier"]:
        cid = cron["identifier"]
    else:
        cid = SALT_CRON_NO_IDENTIFIER
    if cid:
        return _ensure_string(cid)


def _cron_matched(cron, cmd, identifier=None):
    """Check if:
    - we find a cron with same cmd, old state behavior
    - but also be smart enough to remove states changed crons where we do
      not removed priorly by a cron.absent by matching on the provided
      identifier.
      We assure retrocompatibility by only checking on identifier if
      and only if an identifier was set on the serialized crontab
    """
    ret, id_matched = False, None
    cid = _cron_id(cron)
    if cid:
        if not identifier:
            identifier = SALT_CRON_NO_IDENTIFIER
        eidentifier = _ensure_string(identifier)
        # old style second round
        # after saving crontab, we must check that if
        # we have not the same command, but the default id
        # to not set that as a match
        if (
            cron.get("cmd", None) != cmd
            and cid == SALT_CRON_NO_IDENTIFIER
            and eidentifier == SALT_CRON_NO_IDENTIFIER
        ):
            id_matched = False
        else:
            # on saving, be sure not to overwrite a cron
            # with specific identifier but also track
            # crons where command is the same
            # but with the default if that we gonna overwrite
            if (
                cron.get("cmd", None) == cmd
                and cid == SALT_CRON_NO_IDENTIFIER
                and identifier
            ):
                cid = eidentifier
            id_matched = eidentifier == cid
    if ((id_matched is None) and cmd == cron.get("cmd", None)) or id_matched:
        ret = True
    return ret


def _needs_change(old, new):
    if old != new:
        if new == "random":
            # Allow switch from '*' or not present to 'random'
            if old == "*":
                return True
        elif new is not None:
            return True
    return False


def _render_tab(lst):
    """
    Takes a tab list structure and renders it to a list for applying it to
    a file
    """
    ret = []
    for pre in lst["pre"]:
        ret.append("{}\n".format(pre))
    if ret:
        if ret[-1] != TAG:
            ret.append(TAG)
    else:
        ret.append(TAG)
    for env in lst["env"]:
        if (env["value"] is None) or (env["value"] == ""):
            ret.append('{}=""\n'.format(env["name"]))
        else:
            ret.append("{}={}\n".format(env["name"], env["value"]))
    for cron in lst["crons"]:
        if cron["comment"] is not None or cron["identifier"] is not None:
            comment = "#"
            if cron["comment"]:
                comment += " {}".format(cron["comment"].replace("\n", "\n# "))
            if cron["identifier"]:
                comment += " {}:{}".format(SALT_CRON_IDENTIFIER, cron["identifier"])

            comment += "\n"
            ret.append(comment)
        ret.append(
            "{}{} {} {} {} {} {}\n".format(
                cron["commented"] is True and "#DISABLED#" or "",
                cron["minute"],
                cron["hour"],
                cron["daymonth"],
                cron["month"],
                cron["dayweek"],
                cron["cmd"],
            )
        )
    for cron in lst["special"]:
        if cron["comment"] is not None or cron["identifier"] is not None:
            comment = "#"
            if cron["comment"]:
                comment += " {}".format(cron["comment"].rstrip().replace("\n", "\n# "))
            if cron["identifier"]:
                comment += " {}:{}".format(SALT_CRON_IDENTIFIER, cron["identifier"])

            comment += "\n"
            ret.append(comment)
        ret.append(
            "{}{} {}\n".format(
                cron["commented"] is True and "#DISABLED#" or "",
                cron["spec"],
                cron["cmd"],
            )
        )
    return ret


def _get_cron_cmdstr(path, user=None):
    """
    Returns a format string, to be used to build a crontab command.
    """
    if user:
        cmd = "crontab -u {}".format(user)
    else:
        cmd = "crontab"
    return "{} {}".format(cmd, path)


def _check_instance_uid_match(user):
    """
    Returns true if running instance's UID matches the specified user UID
    """
    return os.geteuid() == __salt__["file.user_to_uid"](user)


def write_cron_file(user, path):
    """
    Writes the contents of a file to a user's crontab

    CLI Example:

    .. code-block:: bash

        salt '*' cron.write_cron_file root /tmp/new_cron

    .. versionchanged:: 2015.8.9

    .. note::

        Some OS' do not support specifying user via the `crontab` command i.e. (Solaris, AIX)
    """
    # Some OS' do not support specifying user via the `crontab` command
    if __grains__.get("os_family") in ("Solaris", "AIX"):
        return (
            __salt__["cmd.retcode"](
                _get_cron_cmdstr(path), runas=user, python_shell=False
            )
            == 0
        )
    # If Salt is running from same user as requested in cron module we don't need any user switch
    elif _check_instance_uid_match(user):
        return __salt__["cmd.retcode"](_get_cron_cmdstr(path), python_shell=False) == 0
    # If Salt is running from root user it could modify any user's crontab
    elif _check_instance_uid_match("root"):
        return (
            __salt__["cmd.retcode"](_get_cron_cmdstr(path, user), python_shell=False)
            == 0
        )
    # Edge cases here, let's try do a runas
    else:
        return (
            __salt__["cmd.retcode"](
                _get_cron_cmdstr(path), runas=user, python_shell=False
            )
            == 0
        )


def write_cron_file_verbose(user, path):
    """
    Writes the contents of a file to a user's crontab and return error message on error

    CLI Example:

    .. code-block:: bash

        salt '*' cron.write_cron_file_verbose root /tmp/new_cron

    .. versionchanged:: 2015.8.9

    .. note::

        Some OS' do not support specifying user via the `crontab` command i.e. (Solaris, AIX)
    """
    # Some OS' do not support specifying user via the `crontab` command
    if __grains__.get("os_family") in ("Solaris", "AIX"):
        return __salt__["cmd.run_all"](
            _get_cron_cmdstr(path), runas=user, python_shell=False
        )
    # If Salt is running from same user as requested in cron module we don't need any user switch
    elif _check_instance_uid_match(user):
        return __salt__["cmd.run_all"](_get_cron_cmdstr(path), python_shell=False)
    # If Salt is running from root user it could modify any user's crontab
    elif _check_instance_uid_match("root"):
        return __salt__["cmd.run_all"](_get_cron_cmdstr(path, user), python_shell=False)
    # Edge cases here, let's try do a runas
    else:
        return __salt__["cmd.run_all"](
            _get_cron_cmdstr(path), runas=user, python_shell=False
        )


def _write_cron_lines(user, lines):
    """
    Takes a list of lines to be committed to a user's crontab and writes it
    """
    lines = [salt.utils.stringutils.to_str(_l) for _l in lines]
    path = salt.utils.files.mkstemp()
    if _check_instance_uid_match("root") or __grains__.get("os_family") in (
        "Solaris",
        "AIX",
    ):
        # In some cases crontab command should be executed as user rather than root
        with salt.utils.files.fpopen(
            path, "w+", uid=__salt__["file.user_to_uid"](user), mode=0o600
        ) as fp_:
            fp_.writelines(lines)
        ret = __salt__["cmd.run_all"](
            _get_cron_cmdstr(path), runas=user, python_shell=False
        )
    else:
        with salt.utils.files.fpopen(path, "w+", mode=0o600) as fp_:
            fp_.writelines(lines)
        ret = __salt__["cmd.run_all"](_get_cron_cmdstr(path), python_shell=False)
    os.remove(path)
    return ret


def _date_time_match(cron, **kwargs):
    """
    Returns true if the minute, hour, etc. params match their counterparts from
    the dict returned from list_tab().
    """
    return all(
        [
            kwargs.get(x) is None
            or cron[x] == str(kwargs[x])
            or (str(kwargs[x]).lower() == "random" and cron[x] != "*")
            for x in ("minute", "hour", "daymonth", "month", "dayweek")
        ]
    )


def raw_cron(user):
    """
    Return the contents of the user's crontab

    CLI Example:

    .. code-block:: bash

        salt '*' cron.raw_cron root
    """
    # Some OS' do not support specifying user via the `crontab` command
    if __grains__.get("os_family") in ("Solaris", "AIX"):
        cmd = "crontab -l"
        # Preserve line endings
        lines = salt.utils.data.decode(
            __salt__["cmd.run_stdout"](
                cmd, runas=user, ignore_retcode=True, rstrip=False, python_shell=False
            )
        ).splitlines(True)
    # If Salt is running from same user as requested in cron module we don't need any user switch
    elif _check_instance_uid_match(user):
        cmd = "crontab -l"
        # Preserve line endings
        lines = salt.utils.data.decode(
            __salt__["cmd.run_stdout"](
                cmd, ignore_retcode=True, rstrip=False, python_shell=False
            )
        ).splitlines(True)
    # If Salt is running from root user it could modify any user's crontab
    elif _check_instance_uid_match("root"):
        cmd = "crontab -u {} -l".format(user)
        # Preserve line endings
        lines = salt.utils.data.decode(
            __salt__["cmd.run_stdout"](
                cmd, ignore_retcode=True, rstrip=False, python_shell=False
            )
        ).splitlines(True)
    # Edge cases here, let's try do a runas
    else:
        cmd = "crontab -l"
        # Preserve line endings
        lines = salt.utils.data.decode(
            __salt__["cmd.run_stdout"](
                cmd, runas=user, ignore_retcode=True, rstrip=False, python_shell=False
            )
        ).splitlines(True)

    if lines and lines[0].startswith(
        "# DO NOT EDIT THIS FILE - edit the master and reinstall."
    ):
        del lines[0:3]
    return "".join(lines)


def list_tab(user):
    """
    Return the contents of the specified user's crontab

    CLI Example:

    .. code-block:: bash

        salt '*' cron.list_tab root
    """
    data = raw_cron(user)
    ret = {"pre": [], "crons": [], "special": [], "env": []}
    flag = False
    comment = None
    identifier = None
    for line in data.splitlines():
        if line == "# Lines below here are managed by Salt, do not edit":
            flag = True
            continue
        if flag:
            commented_cron_job = False
            if line.startswith("#DISABLED#"):
                # It's a commented cron job
                line = line[10:]
                commented_cron_job = True
            if line.startswith("@"):
                # Its a "special" line
                dat = {}
                comps = line.split()
                if len(comps) < 2:
                    # Invalid line
                    continue
                dat["spec"] = comps[0]
                dat["cmd"] = " ".join(comps[1:])
                dat["identifier"] = identifier
                dat["comment"] = comment
                dat["commented"] = False
                if commented_cron_job:
                    dat["commented"] = True
                ret["special"].append(dat)
                identifier = None
                comment = None
                commented_cron_job = False
            elif line.startswith("#"):
                # It's a comment! Catch it!
                comment_line = line.lstrip("# ")

                # load the identifier if any
                if SALT_CRON_IDENTIFIER in comment_line:
                    parts = comment_line.split(SALT_CRON_IDENTIFIER)
                    comment_line = parts[0].rstrip()
                    # skip leading :
                    if len(parts[1]) > 1:
                        identifier = parts[1][1:]

                if comment is None:
                    comment = comment_line
                else:
                    comment += "\n" + comment_line
            elif line.find("=") > 0 and (
                " " not in line or line.index("=") < line.index(" ")
            ):
                # Appears to be a ENV setup line
                comps = line.split("=", 1)
                dat = {}
                dat["name"] = comps[0]
                dat["value"] = comps[1]
                ret["env"].append(dat)
            elif len(line.split(" ")) > 5:
                # Appears to be a standard cron line
                comps = line.split(" ")
                dat = {
                    "minute": comps[0],
                    "hour": comps[1],
                    "daymonth": comps[2],
                    "month": comps[3],
                    "dayweek": comps[4],
                    "identifier": identifier,
                    "cmd": " ".join(comps[5:]),
                    "comment": comment,
                    "commented": False,
                }
                if commented_cron_job:
                    dat["commented"] = True
                ret["crons"].append(dat)
                identifier = None
                comment = None
                commented_cron_job = False
        else:
            ret["pre"].append(line)
    return ret


# For consistency's sake
ls = salt.utils.functools.alias_function(list_tab, "ls")


def get_entry(user, identifier=None, cmd=None):
    """
    Return the specified entry from user's crontab.
    identifier will be used if specified, otherwise will lookup cmd
    Either identifier or cmd should be specified.

    user:
        User's crontab to query

    identifier:
        Search for line with identifier

    cmd:
        Search for cron line with cmd

    CLI Example:

    .. code-block:: bash

        salt '*' cron.get_entry root identifier=task1
    """
    if identifier and cmd:
        log.warning("Both identifier and cmd are specified. Only using identifier.")
        cmd = None

    cron_entries = list_tab(user).get("crons", []) + list_tab(user).get("special", [])

    for cron_entry in cron_entries:
        if identifier and cron_entry.get("identifier") == identifier:
            return cron_entry
        elif cmd and cron_entry.get("cmd") == cmd:
            return cron_entry
    return False


def set_special(user, special, cmd, commented=False, comment=None, identifier=None):
    """
    Set up a special command in the crontab.

    CLI Example:

    .. code-block:: bash

        salt '*' cron.set_special root @hourly 'echo foobar'
    """
    lst = list_tab(user)
    for cron in lst["special"]:
        cid = _cron_id(cron)
        if _cron_matched(cron, cmd, identifier):
            test_setted_id = (
                cron["identifier"] is None
                and SALT_CRON_NO_IDENTIFIER
                or cron["identifier"]
            )
            tests = [
                (cron["comment"], comment),
                (cron["commented"], commented),
                (identifier, test_setted_id),
                (cron["spec"], special),
            ]
            if cid or identifier:
                tests.append((cron["cmd"], cmd))
            if any([_needs_change(x, y) for x, y in tests]):
                rm_special(user, cmd, identifier=cid)

                # Use old values when setting the new job if there was no
                # change needed for a given parameter
                if not _needs_change(cron["spec"], special):
                    special = cron["spec"]
                if not _needs_change(cron["commented"], commented):
                    commented = cron["commented"]
                if not _needs_change(cron["comment"], comment):
                    comment = cron["comment"]
                if not _needs_change(cron["cmd"], cmd):
                    cmd = cron["cmd"]
                    if cid == SALT_CRON_NO_IDENTIFIER:
                        if identifier:
                            cid = identifier
                        if (
                            cid == SALT_CRON_NO_IDENTIFIER
                            and cron["identifier"] is None
                        ):
                            cid = None
                        cron["identifier"] = cid
                if not cid or (cid and not _needs_change(cid, identifier)):
                    identifier = cid
                jret = set_special(
                    user,
                    special,
                    cmd,
                    commented=commented,
                    comment=comment,
                    identifier=identifier,
                )
                if jret == "new":
                    return "updated"
                else:
                    return jret
            return "present"
    cron = {
        "spec": special,
        "cmd": cmd,
        "identifier": identifier,
        "comment": comment,
        "commented": commented,
    }
    lst["special"].append(cron)

    comdat = _write_cron_lines(user, _render_tab(lst))
    if comdat["retcode"]:
        # Failed to commit, return the error
        return comdat["stderr"]
    return "new"


def _get_cron_date_time(**kwargs):
    """
    Returns a dict of date/time values to be used in a cron entry
    """
    # Define ranges (except daymonth, as it depends on the month)
    range_max = {
        "minute": list(list(range(60))),
        "hour": list(list(range(24))),
        "month": list(list(range(1, 13))),
        "dayweek": list(list(range(7))),
    }

    ret = {}
    for param in ("minute", "hour", "month", "dayweek"):
        value = str(kwargs.get(param, "1")).lower()
        if value == "random":
            ret[param] = str(random.sample(range_max[param], 1)[0])
        elif len(value.split(":")) == 2:
            cron_range = sorted(value.split(":"))
            start, end = int(cron_range[0]), int(cron_range[1])
            ret[param] = str(random.randint(start, end))
        else:
            ret[param] = value

    if ret["month"] in "1 3 5 7 8 10 12".split():
        daymonth_max = 31
    elif ret["month"] in "4 6 9 11".split():
        daymonth_max = 30
    else:
        # This catches both '2' and '*'
        daymonth_max = 28

    daymonth = str(kwargs.get("daymonth", "1")).lower()
    if daymonth == "random":
        ret["daymonth"] = str(
            random.sample(list(list(range(1, (daymonth_max + 1)))), 1)[0]
        )
    else:
        ret["daymonth"] = daymonth

    return ret


def set_job(
    user,
    minute,
    hour,
    daymonth,
    month,
    dayweek,
    cmd,
    commented=False,
    comment=None,
    identifier=None,
):
    """
    Sets a cron job up for a specified user.

    CLI Example:

    .. code-block:: bash

        salt '*' cron.set_job root '*' '*' '*' '*' 1 /usr/local/weekly
    """
    # Scrub the types
    minute = str(minute).lower()
    hour = str(hour).lower()
    daymonth = str(daymonth).lower()
    month = str(month).lower()
    dayweek = str(dayweek).lower()
    lst = list_tab(user)
    for cron in lst["crons"]:
        cid = _cron_id(cron)
        if _cron_matched(cron, cmd, identifier):
            test_setted_id = (
                cron["identifier"] is None
                and SALT_CRON_NO_IDENTIFIER
                or cron["identifier"]
            )
            tests = [
                (cron["comment"], comment),
                (cron["commented"], commented),
                (identifier, test_setted_id),
                (cron["minute"], minute),
                (cron["hour"], hour),
                (cron["daymonth"], daymonth),
                (cron["month"], month),
                (cron["dayweek"], dayweek),
            ]
            if cid or identifier:
                tests.append((cron["cmd"], cmd))
            if any([_needs_change(x, y) for x, y in tests]):
                rm_job(user, cmd, identifier=cid)

                # Use old values when setting the new job if there was no
                # change needed for a given parameter
                if not _needs_change(cron["minute"], minute):
                    minute = cron["minute"]
                if not _needs_change(cron["hour"], hour):
                    hour = cron["hour"]
                if not _needs_change(cron["daymonth"], daymonth):
                    daymonth = cron["daymonth"]
                if not _needs_change(cron["month"], month):
                    month = cron["month"]
                if not _needs_change(cron["dayweek"], dayweek):
                    dayweek = cron["dayweek"]
                if not _needs_change(cron["commented"], commented):
                    commented = cron["commented"]
                if not _needs_change(cron["comment"], comment):
                    comment = cron["comment"]
                if not _needs_change(cron["cmd"], cmd):
                    cmd = cron["cmd"]
                    if cid == SALT_CRON_NO_IDENTIFIER:
                        if identifier:
                            cid = identifier
                        if (
                            cid == SALT_CRON_NO_IDENTIFIER
                            and cron["identifier"] is None
                        ):
                            cid = None
                        cron["identifier"] = cid
                if not cid or (cid and not _needs_change(cid, identifier)):
                    identifier = cid
                jret = set_job(
                    user,
                    minute,
                    hour,
                    daymonth,
                    month,
                    dayweek,
                    cmd,
                    commented=commented,
                    comment=comment,
                    identifier=identifier,
                )
                if jret == "new":
                    return "updated"
                else:
                    return jret
            return "present"
    cron = {
        "cmd": cmd,
        "identifier": identifier,
        "comment": comment,
        "commented": commented,
    }
    cron.update(
        _get_cron_date_time(
            minute=minute, hour=hour, daymonth=daymonth, month=month, dayweek=dayweek
        )
    )
    lst["crons"].append(cron)

    comdat = _write_cron_lines(user, _render_tab(lst))
    if comdat["retcode"]:
        # Failed to commit, return the error
        return comdat["stderr"]
    return "new"


def rm_special(user, cmd, special=None, identifier=None):
    """
    Remove a special cron job for a specified user.

    CLI Example:

    .. code-block:: bash

        salt '*' cron.rm_special root /usr/bin/foo
    """
    lst = list_tab(user)
    ret = "absent"
    rm_ = None
    for ind, val in enumerate(lst["special"]):
        if rm_ is not None:
            break
        if _cron_matched(val, cmd, identifier=identifier):
            if special is None:
                # No special param was specified
                rm_ = ind
            else:
                if val["spec"] == special:
                    rm_ = ind
    if rm_ is not None:
        lst["special"].pop(rm_)
        ret = "removed"
    comdat = _write_cron_lines(user, _render_tab(lst))
    if comdat["retcode"]:
        # Failed to commit, return the error
        return comdat["stderr"]
    return ret


def rm_job(
    user,
    cmd,
    minute=None,
    hour=None,
    daymonth=None,
    month=None,
    dayweek=None,
    identifier=None,
):
    """
    Remove a cron job for a specified user. If any of the day/time params are
    specified, the job will only be removed if the specified params match.

    CLI Example:

    .. code-block:: bash

        salt '*' cron.rm_job root /usr/local/weekly
        salt '*' cron.rm_job root /usr/bin/foo dayweek=1
    """
    lst = list_tab(user)
    ret = "absent"
    rm_ = None
    for ind, val in enumerate(lst["crons"]):
        if rm_ is not None:
            break
        if _cron_matched(val, cmd, identifier=identifier):
            if not any(
                [x is not None for x in (minute, hour, daymonth, month, dayweek)]
            ):
                # No date/time params were specified
                rm_ = ind
            else:
                if _date_time_match(
                    val,
                    minute=minute,
                    hour=hour,
                    daymonth=daymonth,
                    month=month,
                    dayweek=dayweek,
                ):
                    rm_ = ind
    if rm_ is not None:
        lst["crons"].pop(rm_)
        ret = "removed"
        comdat = _write_cron_lines(user, _render_tab(lst))
        if comdat["retcode"]:
            # Failed to commit, return the error
            return comdat["stderr"]
    return ret


rm = salt.utils.functools.alias_function(rm_job, "rm")


def set_env(user, name, value=None):
    """
    Set up an environment variable in the crontab.

    CLI Example:

    .. code-block:: bash

        salt '*' cron.set_env root MAILTO user@example.com
    """
    lst = list_tab(user)
    for env in lst["env"]:
        if name == env["name"]:
            if value != env["value"]:
                rm_env(user, name)
                jret = set_env(user, name, value)
                if jret == "new":
                    return "updated"
                else:
                    return jret
            return "present"
    env = {"name": name, "value": value}
    lst["env"].append(env)
    comdat = _write_cron_lines(user, _render_tab(lst))
    if comdat["retcode"]:
        # Failed to commit, return the error
        return comdat["stderr"]
    return "new"


def rm_env(user, name):
    """
    Remove cron environment variable for a specified user.

    CLI Example:

    .. code-block:: bash

        salt '*' cron.rm_env root MAILTO
    """
    lst = list_tab(user)
    ret = "absent"
    rm_ = None
    for ind, val in enumerate(lst["env"]):
        if name == val["name"]:
            rm_ = ind
    if rm_ is not None:
        lst["env"].pop(rm_)
        ret = "removed"
    comdat = _write_cron_lines(user, _render_tab(lst))
    if comdat["retcode"]:
        # Failed to commit, return the error
        return comdat["stderr"]
    return ret



"""
    :codeauthor: Erik Johnson <erik@saltstack.com>
"""

import errno
import logging
import os
import shutil
import stat
import tempfile
import textwrap

import pytest
import salt.ext.tornado.ioloop
import salt.fileserver.gitfs as gitfs
import salt.utils.files
import salt.utils.gitfs
import salt.utils.platform
import salt.utils.win_functions
import salt.utils.yaml
from salt.utils.gitfs import (
    GITPYTHON_MINVER,
    GITPYTHON_VERSION,
    LIBGIT2_MINVER,
    LIBGIT2_VERSION,
    PYGIT2_MINVER,
    PYGIT2_VERSION,
)
from tests.support.helpers import patched_environ
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import patch
from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import TestCase, skipIf

try:
    import pwd  # pylint: disable=unused-import
except ImportError:
    pass


try:
    import git

    # We still need to use GitPython here for temp repo setup, so we do need to
    # actually import it. But we don't need import pygit2 in this module, we
    # can just use the LooseVersion instances imported along with
    # salt.utils.gitfs to check if we have a compatible version.
    HAS_GITPYTHON = GITPYTHON_VERSION >= GITPYTHON_MINVER
except (ImportError, AttributeError):
    HAS_GITPYTHON = False

try:
    HAS_PYGIT2 = PYGIT2_VERSION >= PYGIT2_MINVER and LIBGIT2_VERSION >= LIBGIT2_MINVER
except AttributeError:
    HAS_PYGIT2 = False

log = logging.getLogger(__name__)

UNICODE_FILENAME = "питон.txt"
UNICODE_DIRNAME = UNICODE_ENVNAME = "соль"
TAG_NAME = "mytag"


def _rmtree_error(func, path, excinfo):
    os.chmod(path, stat.S_IWRITE)
    func(path)


def _clear_instance_map():
    try:
        del salt.utils.gitfs.GitFS.instance_map[
            salt.ext.tornado.ioloop.IOLoop.current()
        ]
    except KeyError:
        pass


@skipIf(not HAS_GITPYTHON, "GitPython >= {} required".format(GITPYTHON_MINVER))
class GitfsConfigTestCase(TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        opts = {
            "sock_dir": self.tmp_sock_dir,
            "gitfs_remotes": ["file://" + self.tmp_repo_dir],
            "gitfs_root": "",
            "fileserver_backend": ["gitfs"],
            "gitfs_base": "master",
            "gitfs_fallback": "",
            "fileserver_events": True,
            "transport": "zeromq",
            "gitfs_mountpoint": "",
            "gitfs_saltenv": [],
            "gitfs_saltenv_whitelist": [],
            "gitfs_saltenv_blacklist": [],
            "gitfs_user": "",
            "gitfs_password": "",
            "gitfs_insecure_auth": False,
            "gitfs_privkey": "",
            "gitfs_pubkey": "",
            "gitfs_passphrase": "",
            "gitfs_refspecs": [
                "+refs/heads/*:refs/remotes/origin/*",
                "+refs/tags/*:refs/tags/*",
            ],
            "gitfs_ssl_verify": True,
            "gitfs_disable_saltenv_mapping": False,
            "gitfs_ref_types": ["branch", "tag", "sha"],
            "gitfs_update_interval": 60,
            "__role": "master",
        }
        opts["cachedir"] = self.tmp_cachedir
        opts["sock_dir"] = self.tmp_sock_dir
        return {gitfs: {"__opts__": opts}}

    @classmethod
    def setUpClass(cls):
        # Clear the instance map so that we make sure to create a new instance
        # for this test class.
        _clear_instance_map()
        cls.tmp_repo_dir = os.path.join(RUNTIME_VARS.TMP, "gitfs_root")
        if salt.utils.platform.is_windows():
            cls.tmp_repo_dir = cls.tmp_repo_dir.replace("\\", "/")
        cls.tmp_cachedir = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)
        cls.tmp_sock_dir = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)

    @classmethod
    def tearDownClass(cls):
        """
        Remove the temporary git repository and gitfs cache directory to ensure
        a clean environment for the other test class(es).
        """
        for path in (cls.tmp_cachedir, cls.tmp_sock_dir):
            try:
                shutil.rmtree(path, onerror=_rmtree_error)
            except OSError as exc:
                if exc.errno == errno.EACCES:
                    log.error("Access error removeing file %s", path)
                    continue
                if exc.errno != errno.EEXIST:
                    raise

    def test_per_saltenv_config(self):
        opts_override = textwrap.dedent(
            """
            gitfs_root: salt

            gitfs_saltenv:
              - baz:
                # when loaded, the "salt://" prefix will be removed
                - mountpoint: salt://baz_mountpoint
                - ref: baz_branch
                - root: baz_root

            gitfs_remotes:

              - file://{0}tmp/repo1:
                - saltenv:
                  - foo:
                    - ref: foo_branch
                    - root: foo_root

              - file://{0}tmp/repo2:
                - mountpoint: repo2
                - saltenv:
                  - baz:
                    - mountpoint: abc
        """.format(
                "/" if salt.utils.platform.is_windows() else ""
            )
        )
        with patch.dict(gitfs.__opts__, salt.utils.yaml.safe_load(opts_override)):
            git_fs = salt.utils.gitfs.GitFS(
                gitfs.__opts__,
                gitfs.__opts__["gitfs_remotes"],
                per_remote_overrides=gitfs.PER_REMOTE_OVERRIDES,
                per_remote_only=gitfs.PER_REMOTE_ONLY,
            )

        # repo1 (branch: foo)
        # The mountpoint should take the default (from gitfs_mountpoint), while
        # ref and root should take the per-saltenv params.
        self.assertEqual(git_fs.remotes[0].mountpoint("foo"), "")
        self.assertEqual(git_fs.remotes[0].ref("foo"), "foo_branch")
        self.assertEqual(git_fs.remotes[0].root("foo"), "foo_root")

        # repo1 (branch: bar)
        # The 'bar' branch does not have a per-saltenv configuration set, so
        # each of the below values should fall back to global values.
        self.assertEqual(git_fs.remotes[0].mountpoint("bar"), "")
        self.assertEqual(git_fs.remotes[0].ref("bar"), "bar")
        self.assertEqual(git_fs.remotes[0].root("bar"), "salt")

        # repo1 (branch: baz)
        # The 'baz' branch does not have a per-saltenv configuration set, but
        # it is defined in the gitfs_saltenv parameter, so the values
        # from that parameter should be returned.
        self.assertEqual(git_fs.remotes[0].mountpoint("baz"), "baz_mountpoint")
        self.assertEqual(git_fs.remotes[0].ref("baz"), "baz_branch")
        self.assertEqual(git_fs.remotes[0].root("baz"), "baz_root")

        # repo2 (branch: foo)
        # The mountpoint should take the per-remote mountpoint value of
        # 'repo2', while ref and root should fall back to global values.
        self.assertEqual(git_fs.remotes[1].mountpoint("foo"), "repo2")
        self.assertEqual(git_fs.remotes[1].ref("foo"), "foo")
        self.assertEqual(git_fs.remotes[1].root("foo"), "salt")

        # repo2 (branch: bar)
        # The 'bar' branch does not have a per-saltenv configuration set, so
        # the mountpoint should take the per-remote mountpoint value of
        # 'repo2', while ref and root should fall back to global values.
        self.assertEqual(git_fs.remotes[1].mountpoint("bar"), "repo2")
        self.assertEqual(git_fs.remotes[1].ref("bar"), "bar")
        self.assertEqual(git_fs.remotes[1].root("bar"), "salt")

        # repo2 (branch: baz)
        # The 'baz' branch has the mountpoint configured as a per-saltenv
        # parameter. The other two should take the values defined in
        # gitfs_saltenv.
        self.assertEqual(git_fs.remotes[1].mountpoint("baz"), "abc")
        self.assertEqual(git_fs.remotes[1].ref("baz"), "baz_branch")
        self.assertEqual(git_fs.remotes[1].root("baz"), "baz_root")


LOAD = {"saltenv": "base"}


class GitFSTestFuncs:
    """
    These are where the tests go, so that they can be run using both GitPython
    and pygit2.

    NOTE: The gitfs.update() has to happen AFTER the setUp is called. This is
    because running it inside the setUp will spawn a new singleton, which means
    that tests which need to mock the __opts__ will be too late; the setUp will
    have created a new singleton that will bypass our mocking. To ensure that
    our tests are reliable and correct, we want to make sure that each test
    uses a new gitfs object, allowing different manipulations of the opts to be
    tested.

    Therefore, keep the following in mind:

    1. Each test needs to call gitfs.update() *after* any patching, and
       *before* calling the function being tested.
    2. Do *NOT* move the gitfs.update() into the setUp.
    """

    @pytest.mark.slow_test
    def test_file_list(self):
        gitfs.update()
        ret = gitfs.file_list(LOAD)
        self.assertIn("testfile", ret)
        self.assertIn(UNICODE_FILENAME, ret)
        # This function does not use os.sep, the Salt fileserver uses the
        # forward slash, hence it being explicitly used to join here.
        self.assertIn("/".join((UNICODE_DIRNAME, "foo.txt")), ret)

    @pytest.mark.slow_test
    def test_dir_list(self):
        gitfs.update()
        ret = gitfs.dir_list(LOAD)
        self.assertIn("grail", ret)
        self.assertIn(UNICODE_DIRNAME, ret)

    def test_find_and_serve_file(self):
        with patch.dict(gitfs.__opts__, {"file_buffer_size": 262144}):
            gitfs.update()

            # find_file
            ret = gitfs.find_file("testfile")
            self.assertEqual("testfile", ret["rel"])

            full_path_to_file = salt.utils.path.join(
                gitfs.__opts__["cachedir"], "gitfs", "refs", "base", "testfile"
            )
            self.assertEqual(full_path_to_file, ret["path"])

            # serve_file
            load = {"saltenv": "base", "path": full_path_to_file, "loc": 0}
            fnd = {"path": full_path_to_file, "rel": "testfile"}
            ret = gitfs.serve_file(load, fnd)

            with salt.utils.files.fopen(
                os.path.join(RUNTIME_VARS.BASE_FILES, "testfile"), "r"
            ) as fp_:  # NB: Why not 'rb'?
                data = fp_.read()

            self.assertDictEqual(ret, {"data": data, "dest": "testfile"})

    def test_file_list_fallback(self):
        with patch.dict(gitfs.__opts__, {"gitfs_fallback": "master"}):
            gitfs.update()
            ret = gitfs.file_list({"saltenv": "notexisting"})
            self.assertIn("testfile", ret)
            self.assertIn(UNICODE_FILENAME, ret)
            # This function does not use os.sep, the Salt fileserver uses the
            # forward slash, hence it being explicitly used to join here.
            self.assertIn("/".join((UNICODE_DIRNAME, "foo.txt")), ret)

    def test_dir_list_fallback(self):
        with patch.dict(gitfs.__opts__, {"gitfs_fallback": "master"}):
            gitfs.update()
            ret = gitfs.dir_list({"saltenv": "notexisting"})
            self.assertIn("grail", ret)
            self.assertIn(UNICODE_DIRNAME, ret)

    def test_find_and_serve_file_fallback(self):
        with patch.dict(
            gitfs.__opts__, {"file_buffer_size": 262144, "gitfs_fallback": "master"}
        ):
            gitfs.update()

            # find_file
            ret = gitfs.find_file("testfile", tgt_env="notexisting")
            self.assertEqual("testfile", ret["rel"])

            full_path_to_file = salt.utils.path.join(
                gitfs.__opts__["cachedir"], "gitfs", "refs", "notexisting", "testfile"
            )
            self.assertEqual(full_path_to_file, ret["path"])

            # serve_file
            load = {"saltenv": "notexisting", "path": full_path_to_file, "loc": 0}
            fnd = {"path": full_path_to_file, "rel": "testfile"}
            ret = gitfs.serve_file(load, fnd)

            with salt.utils.files.fopen(
                os.path.join(RUNTIME_VARS.BASE_FILES, "testfile"), "r"
            ) as fp_:  # NB: Why not 'rb'?
                data = fp_.read()

            self.assertDictEqual(ret, {"data": data, "dest": "testfile"})

    @pytest.mark.slow_test
    def test_envs(self):
        gitfs.update()
        ret = gitfs.envs(ignore_cache=True)
        self.assertIn("base", ret)
        self.assertIn(UNICODE_ENVNAME, ret)
        self.assertIn(TAG_NAME, ret)

    @pytest.mark.slow_test
    def test_ref_types_global(self):
        """
        Test the global gitfs_ref_types config option
        """
        with patch.dict(gitfs.__opts__, {"gitfs_ref_types": ["branch"]}):
            gitfs.update()
            ret = gitfs.envs(ignore_cache=True)
            # Since we are restricting to branches only, the tag should not
            # appear in the envs list.
            self.assertIn("base", ret)
            self.assertIn(UNICODE_ENVNAME, ret)
            self.assertNotIn(TAG_NAME, ret)

    @pytest.mark.slow_test
    def test_ref_types_per_remote(self):
        """
        Test the per_remote ref_types config option, using a different
        ref_types setting than the global test.
        """
        remotes = [{"file://" + self.tmp_repo_dir: [{"ref_types": ["tag"]}]}]
        with patch.dict(gitfs.__opts__, {"gitfs_remotes": remotes}):
            gitfs.update()
            ret = gitfs.envs(ignore_cache=True)
            # Since we are restricting to tags only, the tag should appear in
            # the envs list, but the branches should not.
            self.assertNotIn("base", ret)
            self.assertNotIn(UNICODE_ENVNAME, ret)
            self.assertIn(TAG_NAME, ret)

    @pytest.mark.slow_test
    def test_disable_saltenv_mapping_global_with_mapping_defined_globally(self):
        """
        Test the global gitfs_disable_saltenv_mapping config option, combined
        with the per-saltenv mapping being defined in the global gitfs_saltenv
        option.
        """
        opts = salt.utils.yaml.safe_load(
            textwrap.dedent(
                """\
            gitfs_disable_saltenv_mapping: True
            gitfs_saltenv:
              - foo:
                - ref: somebranch
            """
            )
        )
        with patch.dict(gitfs.__opts__, opts):
            gitfs.update()
            ret = gitfs.envs(ignore_cache=True)
            # Since we are restricting to tags only, the tag should appear in
            # the envs list, but the branches should not.
            self.assertEqual(ret, ["base", "foo"])

    @pytest.mark.slow_test
    def test_saltenv_blacklist(self):
        """
        test saltenv_blacklist
        """
        opts = salt.utils.yaml.safe_load(
            textwrap.dedent(
                """\
            gitfs_saltenv_blacklist: base
            """
            )
        )
        with patch.dict(gitfs.__opts__, opts):
            gitfs.update()
            ret = gitfs.envs(ignore_cache=True)
            assert "base" not in ret
            assert UNICODE_ENVNAME in ret
            assert "mytag" in ret

    @pytest.mark.slow_test
    def test_saltenv_whitelist(self):
        """
        test saltenv_whitelist
        """
        opts = salt.utils.yaml.safe_load(
            textwrap.dedent(
                """\
            gitfs_saltenv_whitelist: base
            """
            )
        )
        with patch.dict(gitfs.__opts__, opts):
            gitfs.update()
            ret = gitfs.envs(ignore_cache=True)
            assert "base" in ret
            assert UNICODE_ENVNAME not in ret
            assert "mytag" not in ret

    @pytest.mark.slow_test
    def test_env_deprecated_opts(self):
        """
        ensure deprecated options gitfs_env_whitelist
        and gitfs_env_blacklist do not cause gitfs to
        not load.
        """
        opts = salt.utils.yaml.safe_load(
            textwrap.dedent(
                """\
            gitfs_env_whitelist: base
            gitfs_env_blacklist: ''
            """
            )
        )
        with patch.dict(gitfs.__opts__, opts):
            gitfs.update()
            ret = gitfs.envs(ignore_cache=True)
            assert "base" in ret
            assert UNICODE_ENVNAME in ret
            assert "mytag" in ret

    @pytest.mark.slow_test
    def test_disable_saltenv_mapping_global_with_mapping_defined_per_remote(self):
        """
        Test the global gitfs_disable_saltenv_mapping config option, combined
        with the per-saltenv mapping being defined in the remote itself via the
        "saltenv" per-remote option.
        """
        opts = salt.utils.yaml.safe_load(
            textwrap.dedent(
                """\
            gitfs_disable_saltenv_mapping: True
            gitfs_remotes:
              - {}:
                - saltenv:
                  - bar:
                    - ref: somebranch
            """.format(
                    self.tmp_repo_dir
                )
            )
        )
        with patch.dict(gitfs.__opts__, opts):
            gitfs.update()
            ret = gitfs.envs(ignore_cache=True)
            # Since we are restricting to tags only, the tag should appear in
            # the envs list, but the branches should not.
            self.assertEqual(ret, ["bar", "base"])

    @pytest.mark.slow_test
    def test_disable_saltenv_mapping_per_remote_with_mapping_defined_globally(self):
        """
        Test the per-remote disable_saltenv_mapping config option, combined
        with the per-saltenv mapping being defined in the global gitfs_saltenv
        option.
        """
        opts = salt.utils.yaml.safe_load(
            textwrap.dedent(
                """\
            gitfs_remotes:
              - {}:
                - disable_saltenv_mapping: True

            gitfs_saltenv:
              - hello:
                - ref: somebranch
            """.format(
                    self.tmp_repo_dir
                )
            )
        )
        with patch.dict(gitfs.__opts__, opts):
            gitfs.update()
            ret = gitfs.envs(ignore_cache=True)
            # Since we are restricting to tags only, the tag should appear in
            # the envs list, but the branches should not.
            self.assertEqual(ret, ["base", "hello"])

    @pytest.mark.slow_test
    def test_disable_saltenv_mapping_per_remote_with_mapping_defined_per_remote(self):
        """
        Test the per-remote disable_saltenv_mapping config option, combined
        with the per-saltenv mapping being defined in the remote itself via the
        "saltenv" per-remote option.
        """
        opts = salt.utils.yaml.safe_load(
            textwrap.dedent(
                """\
            gitfs_remotes:
              - {}:
                - disable_saltenv_mapping: True
                - saltenv:
                  - world:
                    - ref: somebranch
            """.format(
                    self.tmp_repo_dir
                )
            )
        )
        with patch.dict(gitfs.__opts__, opts):
            gitfs.update()
            ret = gitfs.envs(ignore_cache=True)
            # Since we are restricting to tags only, the tag should appear in
            # the envs list, but the branches should not.
            self.assertEqual(ret, ["base", "world"])


class GitFSTestBase:
    @classmethod
    def setUpClass(cls):
        cls.tmp_repo_dir = os.path.join(RUNTIME_VARS.TMP, "gitfs_root")
        if salt.utils.platform.is_windows():
            cls.tmp_repo_dir = cls.tmp_repo_dir.replace("\\", "/")
        cls.tmp_cachedir = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)
        cls.tmp_sock_dir = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)

        try:
            shutil.rmtree(cls.tmp_repo_dir)
        except OSError as exc:
            if exc.errno == errno.EACCES:
                log.error("Access error removing file %s", cls.tmp_repo_dir)
            elif exc.errno != errno.ENOENT:
                raise

        shutil.copytree(
            str(RUNTIME_VARS.BASE_FILES),
            str(cls.tmp_repo_dir + "/"),
            symlinks=True,
        )

        repo = git.Repo.init(cls.tmp_repo_dir)

        try:
            if salt.utils.platform.is_windows():
                username = salt.utils.win_functions.get_current_user()
            else:
                username = pwd.getpwuid(os.geteuid()).pw_name
        except AttributeError:
            log.error("Unable to get effective username, falling back to 'root'.")
            username = "root"

        with patched_environ(USERNAME=username):
            repo.index.add([x for x in os.listdir(cls.tmp_repo_dir) if x != ".git"])
            repo.index.commit("Test")

            # Add another branch with unicode characters in the name
            repo.create_head(UNICODE_ENVNAME, "HEAD")

            # Add a tag
            repo.create_tag(TAG_NAME, "HEAD")
            # Older GitPython versions do not have a close method.
            if hasattr(repo, "close"):
                repo.close()

    @classmethod
    def tearDownClass(cls):
        """
        Remove the temporary git repository and gitfs cache directory to ensure
        a clean environment for the other test class(es).
        """
        for path in (cls.tmp_cachedir, cls.tmp_sock_dir, cls.tmp_repo_dir):
            try:
                salt.utils.files.rm_rf(path)
            except OSError as exc:
                if exc.errno == errno.EACCES:
                    log.error("Access error removeing file %s", path)
                elif exc.errno != errno.EEXIST:
                    raise

    def setUp(self):
        """
        We don't want to check in another .git dir into GH because that just
        gets messy. Instead, we'll create a temporary repo on the fly for the
        tests to examine.

        Also ensure we A) don't re-use the singleton, and B) that the cachedirs
        are cleared. This keeps these performance enhancements from affecting
        the results of subsequent tests.
        """
        if not gitfs.__virtual__():
            self.skipTest("GitFS could not be loaded. Skipping GitFS tests!")

        _clear_instance_map()
        for subdir in ("gitfs", "file_lists"):
            try:
                salt.utils.files.rm_rf(os.path.join(self.tmp_cachedir, subdir))
            except OSError as exc:
                if exc.errno == errno.EACCES:
                    log.warning(
                        "Access error removeing file %s",
                        os.path.join(self.tmp_cachedir, subdir),
                    )
                    continue
                if exc.errno != errno.ENOENT:
                    raise
        if salt.utils.platform.is_windows():
            self.setUpClass()
            self.setup_loader_modules()


@skipIf(not HAS_GITPYTHON, "GitPython >= {} required".format(GITPYTHON_MINVER))
class GitPythonTest(GitFSTestBase, GitFSTestFuncs, TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        opts = {
            "sock_dir": self.tmp_sock_dir,
            "gitfs_remotes": ["file://" + self.tmp_repo_dir],
            "gitfs_root": "",
            "fileserver_backend": ["gitfs"],
            "gitfs_base": "master",
            "gitfs_fallback": "",
            "fileserver_events": True,
            "transport": "zeromq",
            "gitfs_mountpoint": "",
            "gitfs_saltenv": [],
            "gitfs_saltenv_whitelist": [],
            "gitfs_saltenv_blacklist": [],
            "gitfs_user": "",
            "gitfs_password": "",
            "gitfs_insecure_auth": False,
            "gitfs_privkey": "",
            "gitfs_pubkey": "",
            "gitfs_passphrase": "",
            "gitfs_refspecs": [
                "+refs/heads/*:refs/remotes/origin/*",
                "+refs/tags/*:refs/tags/*",
            ],
            "gitfs_ssl_verify": True,
            "gitfs_disable_saltenv_mapping": False,
            "gitfs_ref_types": ["branch", "tag", "sha"],
            "gitfs_update_interval": 60,
            "__role": "master",
        }
        opts["cachedir"] = self.tmp_cachedir
        opts["sock_dir"] = self.tmp_sock_dir
        opts["gitfs_provider"] = "gitpython"
        return {gitfs: {"__opts__": opts}}


@skipIf(
    not HAS_GITPYTHON,
    "GitPython >= {} required for temp repo setup".format(GITPYTHON_MINVER),
)
@skipIf(
    not HAS_PYGIT2,
    "pygit2 >= {} and libgit2 >= {} required".format(PYGIT2_MINVER, LIBGIT2_MINVER),
)
@skipIf(
    salt.utils.platform.is_windows(),
    "Skip Pygit2 on windows, due to pygit2 access error on windows",
)
class Pygit2Test(GitFSTestBase, GitFSTestFuncs, TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        opts = {
            "sock_dir": self.tmp_sock_dir,
            "gitfs_remotes": ["file://" + self.tmp_repo_dir],
            "gitfs_root": "",
            "fileserver_backend": ["gitfs"],
            "gitfs_base": "master",
            "gitfs_fallback": "",
            "fileserver_events": True,
            "transport": "zeromq",
            "gitfs_mountpoint": "",
            "gitfs_saltenv": [],
            "gitfs_saltenv_whitelist": [],
            "gitfs_saltenv_blacklist": [],
            "gitfs_user": "",
            "gitfs_password": "",
            "gitfs_insecure_auth": False,
            "gitfs_privkey": "",
            "gitfs_pubkey": "",
            "gitfs_passphrase": "",
            "gitfs_refspecs": [
                "+refs/heads/*:refs/remotes/origin/*",
                "+refs/tags/*:refs/tags/*",
            ],
            "gitfs_ssl_verify": True,
            "gitfs_disable_saltenv_mapping": False,
            "gitfs_ref_types": ["branch", "tag", "sha"],
            "gitfs_update_interval": 60,
            "__role": "master",
        }
        opts["cachedir"] = self.tmp_cachedir
        opts["sock_dir"] = self.tmp_sock_dir
        opts["gitfs_provider"] = "pygit2"
        return {gitfs: {"__opts__": opts}}



"""
    :codeauthor: Gareth J. Greenaway <gareth@saltstack.com>
"""

import copy
import logging

import pytest
import salt.proxy.netmiko_px as netmiko_proxy
from saltfactories.utils import random_string
from tests.support.mock import MagicMock, patch

log = logging.getLogger(__name__)


@pytest.fixture(scope="module")
def proxy_minion_config_module(salt_master_factory):
    factory = salt_master_factory.salt_proxy_minion_daemon(
        random_string("proxy-minion-"),
    )
    return factory.config


@pytest.fixture
def proxy_minion_config(proxy_minion_config_module):
    return copy.deepcopy(proxy_minion_config_module)


@pytest.fixture
def configure_loader_modules():
    return {netmiko_proxy: {}}


class MockNetmikoConnection:
    def is_alive(self):
        return False

    def send_config_set(self, *args, **kwargs):
        return args, kwargs


def test_check_virtual():
    """
    check netmiko_proxy virtual method - return value
    """

    with patch.object(netmiko_proxy, "HAS_NETMIKO", True):
        result = netmiko_proxy.__virtual__()
        assert "netmiko" in result

    expected = (
        False,
        "The netmiko proxy module requires netmiko library to be installed.",
    )
    with patch.object(netmiko_proxy, "HAS_NETMIKO", False):
        result = netmiko_proxy.__virtual__()
        assert expected == result


def test_init_skip_connect_on_init_true(proxy_minion_config):
    """
    check netmiko_proxy init method
    """

    proxy_minion_config["skip_connect_on_init"] = True

    assert netmiko_proxy.init(proxy_minion_config)
    assert "netmiko_device" in netmiko_proxy.__context__
    netmiko_device = netmiko_proxy.__context__["netmiko_device"]
    assert "args" in netmiko_device

    assert "initialized" in netmiko_device
    assert not netmiko_device["initialized"]

    assert "up" in netmiko_device
    assert netmiko_device["up"]

    assert "always_alive" in netmiko_device
    assert "connection" not in netmiko_device


def test_init_skip_connect_on_init_false(proxy_minion_config):
    """
    check netmiko_proxy init method
    """

    proxy_minion_config["skip_connect_on_init"] = False

    mock_make_con = MagicMock()
    with patch.object(netmiko_proxy, "make_con", mock_make_con):
        assert netmiko_proxy.init(proxy_minion_config) is None

    assert "netmiko_device" in netmiko_proxy.__context__
    netmiko_device = netmiko_proxy.__context__["netmiko_device"]
    assert "args" in netmiko_device

    assert "initialized" in netmiko_device
    assert netmiko_device["initialized"]

    assert "up" in netmiko_device
    assert netmiko_device["up"]

    assert "always_alive" in netmiko_device
    assert "connection" in netmiko_device


def test_make_con(proxy_minion_config):
    """
    check netmiko_proxy make_con method
    """

    proxy_minion_config["skip_connect_on_init"] = True

    netmiko_proxy.init(proxy_minion_config)

    mock_connection = MockNetmikoConnection

    with patch.object(netmiko_proxy, "ConnectHandler", mock_connection, create=True):
        result = netmiko_proxy.make_con()
        assert result is not None


def test_make_con_raise_exception(proxy_minion_config):
    """
    check netmiko_proxy make_con method
    """

    class ExpectedException(Exception):
        """
        Custom exception class expected to raise as a side_effect
        """

    def raise_exception(*arg, **kwarg):
        raise ExpectedException("expected")

    proxy_minion_config["skip_connect_on_init"] = True

    netmiko_proxy.init(proxy_minion_config)

    with patch.object(netmiko_proxy, "DEFAULT_CONNECTION_TIMEOUT", 0), patch.object(
        netmiko_proxy, "ConnectHandler", raise_exception, create=True
    ):
        with pytest.raises(ExpectedException):
            netmiko_proxy.make_con(0)


def test_ping(proxy_minion_config):
    """
    check netmiko_proxy ping method
    """

    proxy_minion_config["skip_connect_on_init"] = True

    netmiko_proxy.init(proxy_minion_config)

    result = netmiko_proxy.ping()
    assert result is True


def test_alive(proxy_minion_config, subtests):
    """
    check netmiko_proxy alive method
    """

    # Always alive False with skip_connect_on_init on True
    # should return alive as True
    with subtests.test("skip_connect_on_init=True, proxy_always_alive=False"):
        proxy_minion_config["skip_connect_on_init"] = True
        proxy_minion_config["proxy_always_alive"] = False

        netmiko_proxy.init(proxy_minion_config)

        result = netmiko_proxy.alive(proxy_minion_config)
        assert result

    # Always alive True with skip_connect_on_init on True
    # should return alive as False
    with subtests.test("skip_connect_on_init=True, proxy_always_alive=True"):
        proxy_minion_config["skip_connect_on_init"] = True
        proxy_minion_config["proxy_always_alive"] = True

        netmiko_proxy.init(proxy_minion_config)

        result = netmiko_proxy.alive(proxy_minion_config)
        assert not result

    # Always alive True with skip_connect_on_init on False
    # should return alive as True
    with subtests.test("skip_connect_on_init=False, proxy_always_alive=True"):
        proxy_minion_config["skip_connect_on_init"] = False
        proxy_minion_config["proxy_always_alive"] = True

        mock_make_con = MagicMock()
        with patch.object(netmiko_proxy, "make_con", mock_make_con):
            netmiko_proxy.init(proxy_minion_config)
            result = netmiko_proxy.alive(proxy_minion_config)
            assert result


def test_initialized(proxy_minion_config):
    """
    check netmiko_proxy alive method
    """

    proxy_minion_config["skip_connect_on_init"] = True

    netmiko_proxy.init(proxy_minion_config)

    result = netmiko_proxy.initialized()
    assert not result

    # Always alive True with skip_connect_on_init on False
    # should return alive as True
    proxy_minion_config["skip_connect_on_init"] = False

    mock_make_con = MagicMock()
    with patch.object(netmiko_proxy, "make_con", mock_make_con):
        netmiko_proxy.init(proxy_minion_config)
        result = netmiko_proxy.initialized()
        assert result



"""
Support for Gentoolkit

"""

import os

HAS_GENTOOLKIT = False

try:
    from gentoolkit.eclean import search, clean, cli, exclude as excludemod

    HAS_GENTOOLKIT = True
except ImportError:
    pass

# Define the module's virtual name
__virtualname__ = "gentoolkit"


def __virtual__():
    """
    Only work on Gentoo systems with gentoolkit installed
    """
    if __grains__["os"] == "Gentoo" and HAS_GENTOOLKIT:
        return __virtualname__
    return (
        False,
        "The gentoolkitmod execution module cannot be loaded: either the system is not"
        " Gentoo or the gentoolkit.eclean python module not available",
    )


def revdep_rebuild(lib=None):
    """
    Fix up broken reverse dependencies

    lib
        Search for reverse dependencies for a particular library rather
        than every library on the system. It can be a full path to a
        library or basic regular expression.

    CLI Example:

    .. code-block:: bash

        salt '*' gentoolkit.revdep_rebuild
    """
    cmd = "revdep-rebuild -i --quiet --no-progress"
    if lib is not None:
        cmd += " --library={}".format(lib)
    return __salt__["cmd.retcode"](cmd, python_shell=False) == 0


def _pretty_size(size):
    """
    Print sizes in a similar fashion as eclean
    """
    units = [" G", " M", " K", " B"]
    while units and size >= 1000:
        size = size / 1024.0
        units.pop()
    return "{}{}".format(round(size, 1), units[-1])


def _parse_exclude(exclude_file):
    """
    Parse an exclude file.

    Returns a dict as defined in gentoolkit.eclean.exclude.parseExcludeFile
    """
    if os.path.isfile(exclude_file):
        exclude = excludemod.parseExcludeFile(exclude_file, lambda x: None)
    else:
        exclude = dict()
    return exclude


def eclean_dist(
    destructive=False,
    package_names=False,
    size_limit=0,
    time_limit=0,
    fetch_restricted=False,
    exclude_file="/etc/eclean/distfiles.exclude",
):
    """
    Clean obsolete portage sources

    destructive
        Only keep minimum for reinstallation

    package_names
        Protect all versions of installed packages. Only meaningful if used
        with destructive=True

    size_limit <size>
        Don't delete distfiles bigger than <size>.
        <size> is a size specification: "10M" is "ten megabytes",
        "200K" is "two hundreds kilobytes", etc. Units are: G, M, K and B.

    time_limit <time>
        Don't delete distfiles files modified since <time>
        <time> is an amount of time: "1y" is "one year", "2w" is
        "two weeks", etc. Units are: y (years), m (months), w (weeks),
        d (days) and h (hours).

    fetch_restricted
        Protect fetch-restricted files. Only meaningful if used with
        destructive=True

    exclude_file
        Path to exclusion file. Default is /etc/eclean/distfiles.exclude
        This is the same default eclean-dist uses. Use None if this file
        exists and you want to ignore.

    Returns a dict containing the cleaned, saved, and deprecated dists:

    .. code-block:: python

        {'cleaned': {<dist file>: <size>},
         'deprecated': {<package>: <dist file>},
         'saved': {<package>: <dist file>},
         'total_cleaned': <size>}

    CLI Example:

    .. code-block:: bash

        salt '*' gentoolkit.eclean_dist destructive=True
    """
    if exclude_file is None:
        exclude = None
    else:
        try:
            exclude = _parse_exclude(exclude_file)
        except excludemod.ParseExcludeFileException as e:
            ret = {e: "Invalid exclusion file: {}".format(exclude_file)}
            return ret

    if time_limit != 0:
        time_limit = cli.parseTime(time_limit)
    if size_limit != 0:
        size_limit = cli.parseSize(size_limit)

    clean_size = 0
    engine = search.DistfilesSearch(lambda x: None)
    clean_me, saved, deprecated = engine.findDistfiles(
        destructive=destructive,
        package_names=package_names,
        size_limit=size_limit,
        time_limit=time_limit,
        fetch_restricted=fetch_restricted,
        exclude=exclude,
    )

    cleaned = dict()

    def _eclean_progress_controller(size, key, *args):
        cleaned[key] = _pretty_size(size)
        return True

    if clean_me:
        cleaner = clean.CleanUp(_eclean_progress_controller)
        clean_size = cleaner.clean_dist(clean_me)

    ret = {
        "cleaned": cleaned,
        "saved": saved,
        "deprecated": deprecated,
        "total_cleaned": _pretty_size(clean_size),
    }
    return ret


def eclean_pkg(
    destructive=False,
    package_names=False,
    time_limit=0,
    exclude_file="/etc/eclean/packages.exclude",
):
    """
    Clean obsolete binary packages

    destructive
        Only keep minimum for reinstallation

    package_names
        Protect all versions of installed packages. Only meaningful if used
        with destructive=True

    time_limit <time>
        Don't delete distfiles files modified since <time>
        <time> is an amount of time: "1y" is "one year", "2w" is
        "two weeks", etc. Units are: y (years), m (months), w (weeks),
        d (days) and h (hours).

    exclude_file
        Path to exclusion file. Default is /etc/eclean/packages.exclude
        This is the same default eclean-pkg uses. Use None if this file
        exists and you want to ignore.

    Returns a dict containing the cleaned binary packages:

    .. code-block:: python

        {'cleaned': {<dist file>: <size>},
         'total_cleaned': <size>}

    CLI Example:

    .. code-block:: bash

        salt '*' gentoolkit.eclean_pkg destructive=True
    """
    if exclude_file is None:
        exclude = None
    else:
        try:
            exclude = _parse_exclude(exclude_file)
        except excludemod.ParseExcludeFileException as e:
            ret = {e: "Invalid exclusion file: {}".format(exclude_file)}
            return ret

    if time_limit != 0:
        time_limit = cli.parseTime(time_limit)

    clean_size = 0
    # findPackages requires one arg, but does nothing with it.
    # So we will just pass None in for the required arg
    clean_me = search.findPackages(
        None,
        destructive=destructive,
        package_names=package_names,
        time_limit=time_limit,
        exclude=exclude,
        pkgdir=search.pkgdir,
    )

    cleaned = dict()

    def _eclean_progress_controller(size, key, *args):
        cleaned[key] = _pretty_size(size)
        return True

    if clean_me:
        cleaner = clean.CleanUp(_eclean_progress_controller)
        clean_size = cleaner.clean_pkgs(clean_me, search.pkgdir)

    ret = {"cleaned": cleaned, "total_cleaned": _pretty_size(clean_size)}
    return ret


def _glsa_list_process_output(output):
    """
    Process output from glsa_check_list into a dict

    Returns a dict containing the glsa id, description, status, and CVEs
    """
    ret = dict()
    for line in output:
        try:
            glsa_id, status, desc = line.split(None, 2)
            if "U" in status:
                status += " Not Affected"
            elif "N" in status:
                status += " Might be Affected"
            elif "A" in status:
                status += " Applied (injected)"
            if "CVE" in desc:
                desc, cves = desc.rsplit(None, 1)
                cves = cves.split(",")
            else:
                cves = list()
            ret[glsa_id] = {"description": desc, "status": status, "CVEs": cves}
        except ValueError:
            pass
    return ret


def glsa_check_list(glsa_list):
    """
    List the status of Gentoo Linux Security Advisories

    glsa_list
         can contain an arbitrary number of GLSA ids, filenames
         containing GLSAs or the special identifiers 'all' and 'affected'

    Returns a dict containing glsa ids with a description, status, and CVEs:

    .. code-block:: python

        {<glsa_id>: {'description': <glsa_description>,
         'status': <glsa status>,
         'CVEs': [<list of CVEs>]}}

    CLI Example:

    .. code-block:: bash

        salt '*' gentoolkit.glsa_check_list 'affected'
    """
    cmd = "glsa-check --quiet --nocolor --cve --list "
    if isinstance(glsa_list, list):
        for glsa in glsa_list:
            cmd += glsa + " "
    elif glsa_list == "all" or glsa_list == "affected":
        cmd += glsa_list

    ret = dict()
    out = __salt__["cmd.run"](cmd, python_shell=False).split("\n")
    ret = _glsa_list_process_output(out)
    return ret



"""
Manage groups on Mac OS 10.7+
"""

import salt.utils.functools
import salt.utils.itertools
import salt.utils.stringutils
from salt.exceptions import CommandExecutionError, SaltInvocationError
from salt.modules.mac_user import _dscl, _flush_dscl_cache

try:
    import grp
except ImportError:
    pass


# Define the module's virtual name
__virtualname__ = "group"


def __virtual__():
    global _dscl, _flush_dscl_cache
    if __grains__.get("kernel") != "Darwin" or __grains__["osrelease_info"] < (10, 7):
        return (
            False,
            "The mac_group execution module cannot be loaded: only available on"
            " Darwin-based systems >= 10.7",
        )
    _dscl = salt.utils.functools.namespaced_function(_dscl, globals())
    _flush_dscl_cache = salt.utils.functools.namespaced_function(
        _flush_dscl_cache, globals()
    )
    return __virtualname__


def add(name, gid=None, **kwargs):
    """
    Add the specified group

    CLI Example:

    .. code-block:: bash

        salt '*' group.add foo 3456
    """
    ### NOTE: **kwargs isn't used here but needs to be included in this
    ### function for compatibility with the group.present state
    if info(name):
        raise CommandExecutionError("Group '{}' already exists".format(name))
    if salt.utils.stringutils.contains_whitespace(name):
        raise SaltInvocationError("Group name cannot contain whitespace")
    if name.startswith("_"):
        raise SaltInvocationError(
            "Salt will not create groups beginning with underscores"
        )
    if gid is not None and not isinstance(gid, int):
        raise SaltInvocationError("gid must be an integer")
    # check if gid is already in use
    gid_list = _list_gids()
    if str(gid) in gid_list:
        raise CommandExecutionError("gid '{}' already exists".format(gid))

    cmd = ["dseditgroup", "-o", "create"]
    if gid:
        cmd.extend(["-i", gid])
    cmd.append(name)
    return __salt__["cmd.retcode"](cmd, python_shell=False) == 0


def _list_gids():
    """
    Return a list of gids in use
    """
    output = __salt__["cmd.run"](
        ["dscacheutil", "-q", "group"], output_loglevel="quiet", python_shell=False
    )
    ret = set()
    for line in salt.utils.itertools.split(output, "\n"):
        if line.startswith("gid:"):
            ret.update(line.split()[1:])
    return sorted(ret)


def delete(name):
    """
    Remove the named group

    CLI Example:

    .. code-block:: bash

        salt '*' group.delete foo
    """
    if salt.utils.stringutils.contains_whitespace(name):
        raise SaltInvocationError("Group name cannot contain whitespace")
    if name.startswith("_"):
        raise SaltInvocationError(
            "Salt will not remove groups beginning with underscores"
        )
    if not info(name):
        return True
    cmd = ["dseditgroup", "-o", "delete", name]
    return __salt__["cmd.retcode"](cmd, python_shell=False) == 0


def adduser(group, name):
    """
    Add a user in the group.

    CLI Example:

    .. code-block:: bash

         salt '*' group.adduser foo bar

    Verifies if a valid username 'bar' as a member of an existing group 'foo',
    if not then adds it.
    """
    cmd = "dscl . -merge /Groups/{} GroupMembership {}".format(group, name)
    return __salt__["cmd.retcode"](cmd) == 0


def deluser(group, name):
    """
    Remove a user from the group

    .. versionadded:: 2016.3.0

    CLI Example:

    .. code-block:: bash

         salt '*' group.deluser foo bar

    Removes a member user 'bar' from a group 'foo'. If group is not present
    then returns True.
    """
    cmd = "dscl . -delete /Groups/{} GroupMembership {}".format(group, name)
    return __salt__["cmd.retcode"](cmd) == 0


def members(name, members_list):
    """
    Replaces members of the group with a provided list.

    .. versionadded:: 2016.3.0

    CLI Example:

    .. code-block:: bash

        salt '*' group.members foo 'user1,user2,user3,...'

    Replaces a membership list for a local group 'foo'.
    """
    retcode = 1
    grp_info = __salt__["group.info"](name)
    if grp_info and name in grp_info["name"]:
        cmd = "/usr/bin/dscl . -delete /Groups/{} GroupMembership".format(name)
        retcode = __salt__["cmd.retcode"](cmd) == 0
        for user in members_list.split(","):
            cmd = "/usr/bin/dscl . -merge /Groups/{} GroupMembership {}".format(
                name, user
            )
            retcode = __salt__["cmd.retcode"](cmd)
            if not retcode == 0:
                break
            # provided list is '': users previously deleted from group
            else:
                retcode = 0

    return retcode == 0


def info(name):
    """
    Return information about a group

    CLI Example:

    .. code-block:: bash

        salt '*' group.info foo
    """
    if salt.utils.stringutils.contains_whitespace(name):
        raise SaltInvocationError("Group name cannot contain whitespace")
    try:
        # getgrnam seems to cache weirdly, so don't use it
        grinfo = next(iter(x for x in grp.getgrall() if x.gr_name == name))
    except StopIteration:
        return {}
    else:
        return _format_info(grinfo)


def _format_info(data):
    """
    Return formatted information in a pretty way.
    """
    return {
        "name": data.gr_name,
        "gid": data.gr_gid,
        "passwd": data.gr_passwd,
        "members": data.gr_mem,
    }


def getent(refresh=False):
    """
    Return info on all groups

    CLI Example:

    .. code-block:: bash

        salt '*' group.getent
    """
    if "group.getent" in __context__ and not refresh:
        return __context__["group.getent"]

    ret = []
    for grinfo in grp.getgrall():
        if not grinfo.gr_name.startswith("_"):
            ret.append(_format_info(grinfo))
    __context__["group.getent"] = ret
    return ret


def chgid(name, gid):
    """
    Change the gid for a named group

    CLI Example:

    .. code-block:: bash

        salt '*' group.chgid foo 4376
    """
    if not isinstance(gid, int):
        raise SaltInvocationError("gid must be an integer")
    pre_gid = __salt__["file.group_to_gid"](name)
    pre_info = info(name)
    if not pre_info:
        raise CommandExecutionError("Group '{}' does not exist".format(name))
    if gid == pre_info["gid"]:
        return True
    cmd = ["dseditgroup", "-o", "edit", "-i", gid, name]
    return __salt__["cmd.retcode"](cmd, python_shell=False) == 0



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""


import logging

import salt.modules.incron as incron
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, call, patch
from tests.support.unit import TestCase

log = logging.getLogger(__name__)


class IncronTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.incron
    """

    def setup_loader_modules(self):
        return {incron: {}}

    # 'write_incron_file' function tests: 1

    def test_write_incron_file(self):
        """
        Test if it writes the contents of a file to a user's crontab
        """
        mock = MagicMock(return_value=0)
        with patch.dict(incron.__salt__, {"cmd.retcode": mock}), patch(
            "salt.modules.incron._get_incron_cmdstr",
            MagicMock(return_value="incrontab"),
        ):
            self.assertTrue(incron.write_incron_file("cybage", "/home/cybage/new_cron"))

    # 'write_cron_file_verbose' function tests: 1

    def test_write_cron_file_verbose(self):
        """
        Test if it writes the contents of a file to a user's crontab and
        return error message on error
        """
        mock = MagicMock(return_value=True)
        with patch.dict(incron.__salt__, {"cmd.run_all": mock}), patch(
            "salt.modules.incron._get_incron_cmdstr",
            MagicMock(return_value="incrontab"),
        ):
            self.assertTrue(
                incron.write_incron_file_verbose("cybage", "/home/cybage/new_cron")
            )

    # 'raw_system_incron' function tests: 1

    def test_raw_system_incron(self):
        """
        Test if it return the contents of the system wide incrontab
        """
        with patch("salt.modules.incron._read_file", MagicMock(return_value="salt")):
            self.assertEqual(incron.raw_system_incron(), "salt")

    # 'raw_incron' function tests: 1

    def test_raw_incron(self):
        """
        Test if it return the contents of the user's incrontab
        """
        mock = MagicMock(return_value="incrontab")
        expected_calls = [
            call(
                "incrontab -l cybage", python_shell=False, rstrip=False, runas="cybage"
            )
        ]

        with patch.dict(incron.__grains__, {"os_family": mock}):
            cmd_run_mock = MagicMock(return_value="salt")
            with patch.dict(incron.__salt__, {"cmd.run_stdout": cmd_run_mock}):
                self.assertEqual(incron.raw_incron("cybage"), "salt")
                cmd_run_mock.assert_has_calls(expected_calls)

                cmd = cmd_run_mock.call_args[0][0]
                self.assertEqual("incrontab -l cybage", cmd)
                self.assertNotIn("-u", cmd)

    # 'list_tab' function tests: 1

    def test_list_tab(self):
        """
        Test if it return the contents of the specified user's incrontab
        """
        mock = MagicMock(return_value="incrontab")
        with patch.dict(incron.__grains__, {"os_family": mock}):
            mock = MagicMock(return_value="salt")
            with patch.dict(incron.__salt__, {"cmd.run_stdout": mock}):
                self.assertDictEqual(
                    incron.list_tab("cybage"), {"pre": ["salt"], "crons": []}
                )

    # 'set_job' function tests: 1

    def test_set_job(self):
        """
        Test if it sets a cron job up for a specified user.
        """
        self.assertEqual(
            incron.set_job(
                "cybage", "/home/cybage", "TO_MODIFY", 'echo "$$ $@ $# $% $&"'
            ),
            "Invalid mask type: TO_MODIFY",
        )

        val = {
            "pre": [],
            "crons": [
                {"path": "/home/cybage", "mask": "IN_MODIFY", "cmd": 'echo "SALT"'}
            ],
        }
        with patch.object(incron, "list_tab", MagicMock(return_value=val)):
            self.assertEqual(
                incron.set_job("cybage", "/home/cybage", "IN_MODIFY", 'echo "SALT"'),
                "present",
            )

        with patch.object(
            incron, "list_tab", MagicMock(return_value={"pre": ["salt"], "crons": []})
        ):
            mock = MagicMock(return_value="incrontab")
            with patch.dict(incron.__grains__, {"os_family": mock}):
                with patch.object(
                    incron,
                    "_write_incron_lines",
                    MagicMock(return_value={"retcode": True, "stderr": "error"}),
                ):
                    self.assertEqual(
                        incron.set_job(
                            "cybage", "/home/cybage", "IN_MODIFY", 'echo "SALT"'
                        ),
                        "error",
                    )

        with patch.object(
            incron, "list_tab", MagicMock(return_value={"pre": ["salt"], "crons": []})
        ):
            mock = MagicMock(return_value="incrontab")
            with patch.dict(incron.__grains__, {"os_family": mock}):
                with patch.object(
                    incron,
                    "_write_incron_lines",
                    MagicMock(return_value={"retcode": False, "stderr": "error"}),
                ):
                    self.assertEqual(
                        incron.set_job(
                            "cybage", "/home/cybage", "IN_MODIFY", 'echo "SALT"'
                        ),
                        "new",
                    )

        val = {
            "pre": [],
            "crons": [
                {
                    "path": "/home/cybage",
                    "mask": "IN_MODIFY,IN_DELETE",
                    "cmd": 'echo "SALT"',
                }
            ],
        }
        with patch.object(incron, "list_tab", MagicMock(return_value=val)):
            mock = MagicMock(return_value="incrontab")
            with patch.dict(incron.__grains__, {"os_family": mock}):
                with patch.object(
                    incron,
                    "_write_incron_lines",
                    MagicMock(return_value={"retcode": False, "stderr": "error"}),
                ):
                    self.assertEqual(
                        incron.set_job(
                            "cybage", "/home/cybage", "IN_DELETE", 'echo "SALT"'
                        ),
                        "updated",
                    )

    # 'rm_job' function tests: 1

    def test_rm_job(self):
        """
        Test if it remove a cron job for a specified user. If any of the
        day/time params are specified, the job will only be removed if
        the specified params match.
        """
        self.assertEqual(
            incron.rm_job(
                "cybage", "/home/cybage", "TO_MODIFY", 'echo "$$ $@ $# $% $&"'
            ),
            "Invalid mask type: TO_MODIFY",
        )

        with patch.object(
            incron, "list_tab", MagicMock(return_value={"pre": ["salt"], "crons": []})
        ):
            mock = MagicMock(return_value="incrontab")
            with patch.dict(incron.__grains__, {"os_family": mock}):
                with patch.object(
                    incron,
                    "_write_incron_lines",
                    MagicMock(return_value={"retcode": True, "stderr": "error"}),
                ):
                    self.assertEqual(
                        incron.rm_job(
                            "cybage", "/home/cybage", "IN_MODIFY", 'echo "SALT"'
                        ),
                        "error",
                    )

        with patch.object(
            incron, "list_tab", MagicMock(return_value={"pre": ["salt"], "crons": []})
        ):
            mock = MagicMock(return_value="incrontab")
            with patch.dict(incron.__grains__, {"os_family": mock}):
                with patch.object(
                    incron,
                    "_write_incron_lines",
                    MagicMock(return_value={"retcode": False, "stderr": "error"}),
                ):
                    self.assertEqual(
                        incron.rm_job(
                            "cybage", "/home/cybage", "IN_MODIFY", 'echo "SALT"'
                        ),
                        "absent",
                    )



"""
    tests.support.events
    ~~~~~~~~~~~~~~~~~~~~
"""


import multiprocessing
import os
import time
from contextlib import contextmanager

import salt.utils.event
from salt.utils.process import clean_proc


@contextmanager
def eventpublisher_process(sock_dir):
    proc = salt.utils.event.EventPublisher({"sock_dir": sock_dir})
    proc.start()
    try:
        if os.environ.get("TRAVIS_PYTHON_VERSION", None) is not None:
            # Travis is slow
            time.sleep(10)
        else:
            time.sleep(8)
        yield
    finally:
        clean_proc(proc)


class EventSender(multiprocessing.Process):
    def __init__(self, data, tag, wait, sock_dir):
        super().__init__()
        self.data = data
        self.tag = tag
        self.wait = wait
        self.sock_dir = sock_dir

    def run(self):
        with salt.utils.event.MasterEvent(self.sock_dir, listen=False) as me:
            time.sleep(self.wait)
            me.fire_event(self.data, self.tag)
            # Wait a few seconds before tearing down the zmq context
            if os.environ.get("TRAVIS_PYTHON_VERSION", None) is not None:
                # Travis is slow
                time.sleep(10)
            else:
                time.sleep(2)


@contextmanager
def eventsender_process(data, tag, sock_dir, wait=0):
    proc = EventSender(data, tag, wait, sock_dir=sock_dir)
    proc.start()
    try:
        yield
    finally:
        clean_proc(proc)



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import os

import pytest
import salt.modules.btrfs as btrfs
import salt.utils.files
import salt.utils.fsutils
import salt.utils.platform
from salt.exceptions import CommandExecutionError
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, mock_open, patch
from tests.support.unit import TestCase


class BtrfsTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.btrfs
    """

    def setup_loader_modules(self):
        return {btrfs: {"__salt__": {}}}

    # 'version' function tests: 1
    def test_version(self):
        """
        Test if it return BTRFS version.
        """
        mock = MagicMock(return_value={"retcode": 1, "stderr": "", "stdout": "Salt"})
        with patch.dict(btrfs.__salt__, {"cmd.run_all": mock}):
            self.assertDictEqual(btrfs.version(), {"version": "Salt"})

    # 'info' function tests: 1

    def test_info(self):
        """
        Test if it get BTRFS filesystem information.
        """
        with patch("salt.utils.fsutils._verify_run", MagicMock(return_value=True)):
            mock = MagicMock(
                return_value={"retcode": 1, "stderr": "", "stdout": "Salt"}
            )
            with patch.dict(btrfs.__salt__, {"cmd.run_all": mock}):
                mock = MagicMock(return_value={"Salt": "salt"})
                with patch.object(btrfs, "_parse_btrfs_info", mock):
                    self.assertDictEqual(btrfs.info("/dev/sda1"), {"Salt": "salt"})

    # 'devices' function tests: 1

    def test_devices(self):
        """
        Test if it get known BTRFS formatted devices on the system.
        """
        with patch("salt.utils.fsutils._blkid_output", MagicMock(return_value="Salt")):
            mock = MagicMock(
                return_value={"retcode": 1, "stderr": "", "stdout": "Salt"}
            )
            with patch.dict(btrfs.__salt__, {"cmd.run_all": mock}):
                self.assertEqual(btrfs.devices(), "Salt")

    # 'defragment' function tests: 2

    def test_defragment(self):
        """
        Test if it defragment mounted BTRFS filesystem.
        """
        with patch("salt.utils.fsutils._is_device", MagicMock(return_value=False)):
            with patch("os.path.exists", MagicMock(return_value=True)):
                ret = [
                    {
                        "range": "/dev/sda1",
                        "mount_point": False,
                        "log": False,
                        "passed": True,
                    }
                ]
                mock_run = MagicMock(
                    return_value={"retcode": 1, "stderr": "", "stdout": "Salt"}
                )
                with patch.dict(btrfs.__salt__, {"cmd.run_all": mock_run}):
                    mock_file = mock_open(
                        read_data="/dev/sda1 / ext4 rw,data=ordered 0 0"
                    )
                    with patch.object(salt.utils.files, "fopen", mock_file):
                        self.assertListEqual(btrfs.defragment("/dev/sda1"), ret)

    def test_defragment_error(self):
        """
        Test if it gives device not mount error
        """
        with patch("salt.utils.fsutils._is_device", MagicMock(return_value=True)):
            mock_run = MagicMock(
                return_value={"retcode": 1, "stderr": "", "stdout": "Salt"}
            )
            with patch.dict(btrfs.__salt__, {"cmd.run_all": mock_run}):
                mock_file = mock_open(read_data="/dev/sda1 / ext4 rw,data=ordered 0 0")
                with patch.object(salt.utils.files, "fopen", mock_file):
                    self.assertRaises(
                        CommandExecutionError, btrfs.defragment, "/dev/sda1"
                    )

    # 'features' function tests: 1

    def test_features(self):
        """
        Test if it list currently available BTRFS features.
        """
        with patch("salt.utils.fsutils._verify_run", MagicMock(return_value=True)):
            mock = MagicMock(
                return_value={"retcode": 1, "stderr": "", "stdout": "Salt"}
            )
            with patch.dict(btrfs.__salt__, {"cmd.run_all": mock}):
                self.assertDictEqual(btrfs.features(), {})

    # 'usage' function tests: 1

    def test_usage(self):
        """
        Test if it shows in which disk the chunks are allocated.
        """
        with patch("salt.utils.fsutils._verify_run", MagicMock(return_value=True)):
            mock = MagicMock(
                return_value={"retcode": 1, "stderr": "", "stdout": "Salt"}
            )
            with patch.dict(btrfs.__salt__, {"cmd.run_all": mock}):
                mock = MagicMock(return_value={"Salt": "salt"})
                with patch.object(btrfs, "_usage_specific", mock):
                    self.assertDictEqual(btrfs.usage("/dev/sda1"), {"Salt": "salt"})

            mock = MagicMock(
                return_value={"retcode": 1, "stderr": "", "stdout": "Unallocated:\n"}
            )
            with patch.dict(btrfs.__salt__, {"cmd.run_all": mock}):
                mock = MagicMock(return_value={"/dev/sda1": True})
                with patch.object(btrfs, "_usage_unallocated", mock):
                    self.assertDictEqual(
                        btrfs.usage("/dev/sda1"), {"unallocated": {"/dev/sda1": True}}
                    )

            mock = MagicMock(
                return_value={"retcode": 1, "stderr": "", "stdout": "Overall:\n"}
            )
            with patch.dict(btrfs.__salt__, {"cmd.run_all": mock}):
                mock = MagicMock(return_value={"/dev/sda1": True})
                with patch.object(btrfs, "_usage_overall", mock):
                    self.assertDictEqual(
                        btrfs.usage("/dev/sda1"), {"overall": {"/dev/sda1": True}}
                    )

    # 'mkfs' function tests: 3

    def test_mkfs(self):
        """
        Test if it create a file system on the specified device.
        """
        mock_cmd = MagicMock(
            return_value={"retcode": 1, "stderr": "", "stdout": "Salt"}
        )
        mock_info = MagicMock(return_value=[])
        with patch.dict(
            btrfs.__salt__, {"cmd.run_all": mock_cmd, "btrfs.info": mock_info}
        ):
            mock_file = mock_open(read_data="/dev/sda1 / ext4 rw,data=ordered 0 0")
            with patch.object(salt.utils.files, "fopen", mock_file):
                self.assertDictEqual(btrfs.mkfs("/dev/sda1"), {"log": "Salt"})

    def test_mkfs_error(self):
        """
        Test if it No devices specified error
        """
        self.assertRaises(CommandExecutionError, btrfs.mkfs)

    def test_mkfs_mount_error(self):
        """
        Test if it device mount error
        """
        mock = MagicMock(return_value={"/dev/sda1": True})
        with patch.object(salt.utils.fsutils, "_get_mounts", mock):
            self.assertRaises(CommandExecutionError, btrfs.mkfs, "/dev/sda1")

    # 'resize' function tests: 4

    def test_resize(self):
        """
        Test if it resize filesystem.
        """
        with patch("salt.utils.fsutils._is_device", MagicMock(return_value=True)):
            mock = MagicMock(
                return_value={"retcode": 1, "stderr": "", "stdout": "Salt"}
            )
            mock_info = MagicMock(return_value=[])
            with patch.dict(
                btrfs.__salt__, {"cmd.run_all": mock, "btrfs.info": mock_info}
            ):
                mock = MagicMock(return_value={"/dev/sda1": True})
                with patch.object(salt.utils.fsutils, "_get_mounts", mock):
                    self.assertDictEqual(
                        btrfs.resize("/dev/sda1", "max"), {"log": "Salt"}
                    )

    def test_resize_valid_error(self):
        """
        Test if it gives device should be mounted error
        """
        with patch("salt.utils.fsutils._is_device", MagicMock(return_value=False)):
            mock = MagicMock(
                return_value={"retcode": 1, "stderr": "", "stdout": "Salt"}
            )
            with patch.dict(btrfs.__salt__, {"cmd.run_all": mock}):
                self.assertRaises(
                    CommandExecutionError, btrfs.resize, "/dev/sda1", "max"
                )

    def test_resize_mount_error(self):
        """
        Test if it gives mount point error
        """
        with patch("salt.utils.fsutils._is_device", MagicMock(return_value=True)):
            mock = MagicMock(return_value={"/dev/sda1": False})
            with patch.object(salt.utils.fsutils, "_get_mounts", mock):
                self.assertRaises(
                    CommandExecutionError, btrfs.resize, "/dev/sda1", "max"
                )

    def test_resize_size_error(self):
        """
        Test if it gives unknown size error
        """
        self.assertRaises(CommandExecutionError, btrfs.resize, "/dev/sda1", "250m")

    # 'convert' function tests: 5

    def test_convert(self):
        """
        Test if it convert ext2/3/4 to BTRFS
        """
        with patch("os.path.exists", MagicMock(return_value=True)):
            ret = {
                "after": {
                    "balance_log": "Salt",
                    "ext4_image": "removed",
                    "ext4_image_info": "N/A",
                    "fsck_status": "N/A",
                    "mount_point": None,
                    "type": "ext4",
                },
                "before": {
                    "fsck_status": "Filesystem errors corrected",
                    "mount_point": None,
                    "type": "ext4",
                },
            }
            mock = MagicMock(
                return_value={"retcode": 1, "stderr": "", "stdout": "Salt"}
            )
            with patch.dict(btrfs.__salt__, {"cmd.run_all": mock}):
                mock = MagicMock(return_value={"/dev/sda3": {"type": "ext4"}})
                with patch.object(salt.utils.fsutils, "_blkid_output", mock):
                    mock = MagicMock(
                        return_value={"/dev/sda3": [{"mount_point": None}]}
                    )
                    with patch.object(salt.utils.fsutils, "_get_mounts", mock):
                        self.assertDictEqual(
                            btrfs.convert("/dev/sda3", permanent=True), ret
                        )

    def test_convert_device_error(self):
        """
        Test if it gives device not found error
        """
        mock = MagicMock(return_value={"retcode": 1, "stderr": "", "stdout": "Salt"})
        with patch.dict(btrfs.__salt__, {"cmd.run_all": mock}):
            mock = MagicMock(return_value={"/dev/sda1": False})
            with patch.object(salt.utils.fsutils, "_blkid_output", mock):
                self.assertRaises(CommandExecutionError, btrfs.convert, "/dev/sda1")

    def test_convert_filesystem_error(self):
        """
        Test if it gives file system error
        """
        with patch("salt.utils.fsutils._is_device", MagicMock(return_value=True)):
            mock = MagicMock(
                return_value={"retcode": 1, "stderr": "", "stdout": "Salt"}
            )
            with patch.dict(btrfs.__salt__, {"cmd.run_all": mock}):
                mock = MagicMock(return_value={"/dev/sda1": {"type": "ext"}})
                with patch.object(salt.utils.fsutils, "_blkid_output", mock):
                    self.assertRaises(CommandExecutionError, btrfs.convert, "/dev/sda1")

    def test_convert_error(self):
        """
        Test if it gives error cannot convert root
        """
        with patch("salt.utils.fsutils._is_device", MagicMock(return_value=True)):
            mock = MagicMock(
                return_value={"retcode": 1, "stderr": "", "stdout": "Salt"}
            )
            with patch.dict(btrfs.__salt__, {"cmd.run_all": mock}):
                mock = MagicMock(
                    return_value={"/dev/sda1": {"type": "ext4", "mount_point": "/"}}
                )
                with patch.object(salt.utils.fsutils, "_blkid_output", mock):
                    mock = MagicMock(return_value={"/dev/sda1": [{"mount_point": "/"}]})
                    with patch.object(salt.utils.fsutils, "_get_mounts", mock):
                        self.assertRaises(
                            CommandExecutionError, btrfs.convert, "/dev/sda1"
                        )

    def test_convert_migration_error(self):
        """
        Test if it gives migration error
        """
        with patch("salt.utils.fsutils._is_device", MagicMock(return_value=True)):
            mock_run = MagicMock(
                return_value={"retcode": 1, "stderr": "", "stdout": "Salt"}
            )
            with patch.dict(btrfs.__salt__, {"cmd.run_all": mock_run}):
                mock_blk = MagicMock(return_value={"/dev/sda1": {"type": "ext4"}})
                with patch.object(salt.utils.fsutils, "_blkid_output", mock_blk):
                    mock_file = mock_open(
                        read_data="/dev/sda1 / ext4 rw,data=ordered 0 0"
                    )
                    with patch.object(salt.utils.files, "fopen", mock_file):
                        self.assertRaises(
                            CommandExecutionError, btrfs.convert, "/dev/sda1"
                        )

    # 'add' function tests: 1

    def test_add(self):
        """
        Test if it add a devices to a BTRFS filesystem.
        """
        with patch("salt.modules.btrfs._restripe", MagicMock(return_value={})):
            self.assertDictEqual(btrfs.add("/mountpoint", "/dev/sda1", "/dev/sda2"), {})

    # 'delete' function tests: 1

    def test_delete(self):
        """
        Test if it delete a devices to a BTRFS filesystem.
        """
        with patch("salt.modules.btrfs._restripe", MagicMock(return_value={})):
            self.assertDictEqual(
                btrfs.delete("/mountpoint", "/dev/sda1", "/dev/sda2"), {}
            )

    # 'properties' function tests: 1

    def test_properties(self):
        """
        Test if list properties for given btrfs object
        """
        with patch("salt.utils.fsutils._verify_run", MagicMock(return_value=True)):
            mock = MagicMock(
                return_value={"retcode": 1, "stderr": "", "stdout": "Salt"}
            )
            with patch.dict(btrfs.__salt__, {"cmd.run_all": mock}):
                self.assertDictEqual(btrfs.properties("/dev/sda1", "subvol"), {})

    def test_properties_unknown_error(self):
        """
        Test if it gives unknown property error
        """
        self.assertRaises(CommandExecutionError, btrfs.properties, "/dev/sda1", "a")

    def test_properties_error(self):
        """
        Test if it gives exception error
        """
        self.assertRaises(
            CommandExecutionError, btrfs.properties, "/dev/sda1", "subvol", True
        )

    def test_subvolume_exists(self):
        """
        Test subvolume_exists
        """
        salt_mock = {
            "cmd.retcode": MagicMock(return_value=0),
        }
        with patch.dict(btrfs.__salt__, salt_mock):
            assert btrfs.subvolume_exists("/mnt/one")

    def test_subvolume_not_exists(self):
        """
        Test subvolume_exists
        """
        salt_mock = {
            "cmd.retcode": MagicMock(return_value=1),
        }
        with patch.dict(btrfs.__salt__, salt_mock):
            assert not btrfs.subvolume_exists("/mnt/nowhere")

    def test_subvolume_create_fails_parameters(self):
        """
        Test btrfs subvolume create
        """
        # Fails when qgroupids is not a list
        with pytest.raises(CommandExecutionError):
            btrfs.subvolume_create("var", qgroupids="1")

    @patch("salt.modules.btrfs.subvolume_exists")
    def test_subvolume_create_already_exists(self, subvolume_exists):
        """
        Test btrfs subvolume create
        """
        subvolume_exists.return_value = True
        assert not btrfs.subvolume_create("var", dest="/mnt")

    @patch("salt.modules.btrfs.subvolume_exists")
    def test_subvolume_create(self, subvolume_exists):
        """
        Test btrfs subvolume create
        """
        subvolume_exists.return_value = False
        salt_mock = {
            "cmd.run_all": MagicMock(return_value={"recode": 0}),
        }
        expected_path = os.path.join("/mnt", "var")
        with patch.dict(btrfs.__salt__, salt_mock):
            assert btrfs.subvolume_create("var", dest="/mnt")
            subvolume_exists.assert_called_once()
            salt_mock["cmd.run_all"].assert_called_once()
            salt_mock["cmd.run_all"].assert_called_with(
                ["btrfs", "subvolume", "create", expected_path]
            )

    def test_subvolume_delete_fails_parameters(self):
        """
        Test btrfs subvolume delete
        """
        # We need to provide name or names
        with pytest.raises(CommandExecutionError):
            btrfs.subvolume_delete()

        with pytest.raises(CommandExecutionError):
            btrfs.subvolume_delete(names="var")

    def test_subvolume_delete_fails_parameter_commit(self):
        """
        Test btrfs subvolume delete
        """
        # Parameter commit can be 'after' or 'each'
        with pytest.raises(CommandExecutionError):
            btrfs.subvolume_delete(name="var", commit="maybe")

    @patch("salt.modules.btrfs.subvolume_exists")
    def test_subvolume_delete_already_missing(self, subvolume_exists):
        """
        Test btrfs subvolume delete
        """
        subvolume_exists.return_value = False
        assert not btrfs.subvolume_delete(name="var", names=["tmp"])

    @patch("salt.modules.btrfs.subvolume_exists")
    def test_subvolume_delete_already_missing_name(self, subvolume_exists):
        """
        Test btrfs subvolume delete
        """
        subvolume_exists.return_value = False
        assert not btrfs.subvolume_delete(name="var")

    @patch("salt.modules.btrfs.subvolume_exists")
    def test_subvolume_delete_already_missing_names(self, subvolume_exists):
        """
        Test btrfs subvolume delete
        """
        subvolume_exists.return_value = False
        assert not btrfs.subvolume_delete(names=["tmp"])

    @patch("salt.modules.btrfs.subvolume_exists")
    def test_subvolume_delete(self, subvolume_exists):
        """
        Test btrfs subvolume delete
        """
        subvolume_exists.return_value = True
        salt_mock = {
            "cmd.run_all": MagicMock(return_value={"recode": 0}),
        }
        with patch.dict(btrfs.__salt__, salt_mock):
            assert btrfs.subvolume_delete("var", names=["tmp"])
            salt_mock["cmd.run_all"].assert_called_once()
            salt_mock["cmd.run_all"].assert_called_with(
                ["btrfs", "subvolume", "delete", "var", "tmp"]
            )

    def test_subvolume_find_new_empty(self):
        """
        Test btrfs subvolume find-new
        """
        salt_mock = {
            "cmd.run_all": MagicMock(
                return_value={"recode": 0, "stdout": "transid marker was 1024"}
            ),
        }
        with patch.dict(btrfs.__salt__, salt_mock):
            assert btrfs.subvolume_find_new("var", "2000") == {
                "files": [],
                "transid": "1024",
            }
            salt_mock["cmd.run_all"].assert_called_once()
            salt_mock["cmd.run_all"].assert_called_with(
                ["btrfs", "subvolume", "find-new", "var", "2000"]
            )

    def test_subvolume_find_new(self):
        """
        Test btrfs subvolume find-new
        """
        salt_mock = {
            "cmd.run_all": MagicMock(
                return_value={
                    "recode": 0,
                    "stdout": """inode 185148 ... gen 2108 flags NONE var/log/audit/audit.log
inode 187390 ... INLINE etc/openvpn/openvpn-status.log
transid marker was 1024""",
                }
            ),
        }
        with patch.dict(btrfs.__salt__, salt_mock):
            assert btrfs.subvolume_find_new("var", "1023") == {
                "files": ["var/log/audit/audit.log", "etc/openvpn/openvpn-status.log"],
                "transid": "1024",
            }
            salt_mock["cmd.run_all"].assert_called_once()
            salt_mock["cmd.run_all"].assert_called_with(
                ["btrfs", "subvolume", "find-new", "var", "1023"]
            )

    def test_subvolume_get_default_free(self):
        """
        Test btrfs subvolume get-default
        """
        salt_mock = {
            "cmd.run_all": MagicMock(
                return_value={"recode": 0, "stdout": "ID 5 (FS_TREE)"}
            ),
        }
        with patch.dict(btrfs.__salt__, salt_mock):
            assert btrfs.subvolume_get_default("/mnt") == {
                "id": "5",
                "name": "(FS_TREE)",
            }
            salt_mock["cmd.run_all"].assert_called_once()
            salt_mock["cmd.run_all"].assert_called_with(
                ["btrfs", "subvolume", "get-default", "/mnt"]
            )

    def test_subvolume_get_default(self):
        """
        Test btrfs subvolume get-default
        """
        salt_mock = {
            "cmd.run_all": MagicMock(
                return_value={
                    "recode": 0,
                    "stdout": "ID 257 gen 8 top level 5 path var",
                }
            ),
        }
        with patch.dict(btrfs.__salt__, salt_mock):
            assert btrfs.subvolume_get_default("/mnt") == {
                "id": "257",
                "name": "var",
            }
            salt_mock["cmd.run_all"].assert_called_once()
            salt_mock["cmd.run_all"].assert_called_with(
                ["btrfs", "subvolume", "get-default", "/mnt"]
            )

    def test_subvolume_list_fails_parameters(self):
        """
        Test btrfs subvolume list
        """
        # Fails when sort is not a list
        with pytest.raises(CommandExecutionError):
            btrfs.subvolume_list("/mnt", sort="-rootid")

        # Fails when sort is not recognized
        with pytest.raises(CommandExecutionError):
            btrfs.subvolume_list("/mnt", sort=["-root"])

    def test_subvolume_list_simple(self):
        """
        Test btrfs subvolume list
        """
        salt_mock = {
            "cmd.run_all": MagicMock(
                return_value={
                    "recode": 0,
                    "stdout": """ID 257 gen 8 top level 5 path one
ID 258 gen 10 top level 5 path another one
""",
                }
            ),
        }
        with patch.dict(btrfs.__salt__, salt_mock):
            assert btrfs.subvolume_list("/mnt") == [
                {"id": "257", "gen": "8", "top level": "5", "path": "one"},
                {"id": "258", "gen": "10", "top level": "5", "path": "another one"},
            ]
            salt_mock["cmd.run_all"].assert_called_once()
            salt_mock["cmd.run_all"].assert_called_with(
                ["btrfs", "subvolume", "list", "/mnt"]
            )

    def test_subvolume_list(self):
        """
        Test btrfs subvolume list
        """
        salt_mock = {
            "cmd.run_all": MagicMock(
                return_value={
                    "recode": 0,
                    "stdout": """\
ID 257 gen 8 cgen 8 parent 5 top level 5 parent_uuid -     received_uuid - \
             uuid 777...-..05 path one
ID 258 gen 10 cgen 10 parent 5 top level 5 parent_uuid -     received_uuid - \
             uuid a90...-..01 path another one
""",
                }
            ),
        }
        with patch.dict(btrfs.__salt__, salt_mock):
            assert btrfs.subvolume_list(
                "/mnt",
                parent_id=True,
                absolute=True,
                ogeneration=True,
                generation=True,
                subvolumes=True,
                uuid=True,
                parent_uuid=True,
                sent_subvolume_uuid=True,
                generation_cmp="-100",
                ogeneration_cmp="+5",
                sort=["-rootid", "gen"],
            ) == [
                {
                    "id": "257",
                    "gen": "8",
                    "cgen": "8",
                    "parent": "5",
                    "top level": "5",
                    "parent_uuid": "-",
                    "received_uuid": "-",
                    "uuid": "777...-..05",
                    "path": "one",
                },
                {
                    "id": "258",
                    "gen": "10",
                    "cgen": "10",
                    "parent": "5",
                    "top level": "5",
                    "parent_uuid": "-",
                    "received_uuid": "-",
                    "uuid": "a90...-..01",
                    "path": "another one",
                },
            ]
            salt_mock["cmd.run_all"].assert_called_once()
            salt_mock["cmd.run_all"].assert_called_with(
                [
                    "btrfs",
                    "subvolume",
                    "list",
                    "-p",
                    "-a",
                    "-c",
                    "-g",
                    "-o",
                    "-u",
                    "-q",
                    "-R",
                    "-G",
                    "-100",
                    "-C",
                    "+5",
                    "--sort=-rootid,gen",
                    "/mnt",
                ]
            )

    def test_subvolume_set_default(self):
        """
        Test btrfs subvolume set-default
        """
        salt_mock = {
            "cmd.run_all": MagicMock(return_value={"recode": 0}),
        }
        with patch.dict(btrfs.__salt__, salt_mock):
            assert btrfs.subvolume_set_default("257", "/mnt")
            salt_mock["cmd.run_all"].assert_called_once()
            salt_mock["cmd.run_all"].assert_called_with(
                ["btrfs", "subvolume", "set-default", "257", "/mnt"]
            )

    def test_subvolume_show(self):
        """
        Test btrfs subvolume show
        """
        salt_mock = {
            "cmd.run_all": MagicMock(
                return_value={
                    "recode": 0,
                    "stdout": """@/var
        Name:                   var
        UUID:                   7a14...-...04
        Parent UUID:            -
        Received UUID:          -
        Creation time:          2018-10-01 14:33:12 +0200
        Subvolume ID:           258
        Generation:             82479
        Gen at creation:        10
        Parent ID:              256
        Top level ID:           256
        Flags:                  -
        Snapshot(s):
""",
                }
            ),
        }
        with patch.dict(btrfs.__salt__, salt_mock):
            assert btrfs.subvolume_show("/var") == {
                "@/var": {
                    "name": "var",
                    "uuid": "7a14...-...04",
                    "parent uuid": "-",
                    "received uuid": "-",
                    "creation time": "2018-10-01 14:33:12 +0200",
                    "subvolume id": "258",
                    "generation": "82479",
                    "gen at creation": "10",
                    "parent id": "256",
                    "top level id": "256",
                    "flags": "-",
                    "snapshot(s)": "",
                },
            }
            salt_mock["cmd.run_all"].assert_called_once()
            salt_mock["cmd.run_all"].assert_called_with(
                ["btrfs", "subvolume", "show", "/var"]
            )

    def test_subvolume_sync_fail_parameters(self):
        """
        Test btrfs subvolume sync
        """
        # Fails when subvolids is not a list
        with pytest.raises(CommandExecutionError):
            btrfs.subvolume_sync("/mnt", subvolids="257")

    def test_subvolume_sync(self):
        """
        Test btrfs subvolume sync
        """
        salt_mock = {
            "cmd.run_all": MagicMock(return_value={"recode": 0}),
        }
        with patch.dict(btrfs.__salt__, salt_mock):
            assert btrfs.subvolume_sync("/mnt", subvolids=["257"], sleep="1")
            salt_mock["cmd.run_all"].assert_called_once()
            salt_mock["cmd.run_all"].assert_called_with(
                ["btrfs", "subvolume", "sync", "-s", "1", "/mnt", "257"]
            )



import pytest
from tests.support.mock import patch

try:
    import salt.states.reg as reg
    import salt.utils.win_dacl as win_dacl
    import salt.utils.win_functions as win_functions
    import salt.utils.win_reg as reg_util

    HAS_WIN_LIBS = True
except ImportError:
    HAS_WIN_LIBS = False

pytestmark = [
    pytest.mark.skip_unless_on_windows,
    pytest.mark.skipif(HAS_WIN_LIBS is False, reason="Windows Libraries not available"),
    pytest.mark.windows_whitelisted,
]


class RegVars:
    """
    Class for tests that require shared variables
    """

    def __init__(self):
        self.hive = "HKEY_CURRENT_USER"
        self.key = "SOFTWARE\\Salt-Testing"
        self.hive = self.hive
        self.key = self.key
        self.name = "{}\\{}".format(self.hive, self.key)
        self.vname = "version"
        self.vdata = "0.15.3"
        self.current_user = win_functions.get_current_user(with_domain=False)


@pytest.fixture(scope="module")
def configure_loader_modules():
    return {
        reg: {
            "__opts__": {"test": False},
            "__salt__": {},
            "__utils__": {
                "reg.cast_vdata": reg_util.cast_vdata,
                "reg.delete_value": reg_util.delete_value,
                "reg.read_value": reg_util.read_value,
                "reg.set_value": reg_util.set_value,
                "dacl.check_perms": win_dacl.check_perms,
            },
        },
        win_dacl: {"__opts__": {"test": False}},
    }


@pytest.fixture(scope="function")
def reg_vars():
    return RegVars()


@pytest.fixture(scope="function")
def clean(reg_vars):
    """
    Make sure the key does NOT exist
    """
    try:
        if reg_util.key_exists(hive=reg_vars.hive, key=reg_vars.key):
            reg_util.delete_key_recursive(hive=reg_vars.hive, key=reg_vars.key)
        yield
    finally:
        if reg_util.key_exists(hive=reg_vars.hive, key=reg_vars.key):
            reg_util.delete_key_recursive(hive=reg_vars.hive, key=reg_vars.key)


@pytest.fixture(scope="function")
def reset(reg_vars):
    """
    Create an existing key for testing
    """
    try:
        if not reg_util.value_exists(
            hive=reg_vars.hive, key=reg_vars.key, vname=reg_vars.vname
        ):
            reg_util.set_value(
                hive=reg_vars.hive,
                key=reg_vars.key,
                vname=reg_vars.vname,
                vdata=reg_vars.vdata,
            )
        yield
    finally:
        if reg_util.key_exists(hive=reg_vars.hive, key=reg_vars.key):
            reg_util.delete_key_recursive(hive=reg_vars.hive, key=reg_vars.key)


def test_present(reg_vars):
    """
    Test reg.present
    """
    expected = {
        "comment": "Added {} to {}".format(reg_vars.vname, reg_vars.name),
        "changes": {
            "reg": {
                "Added": {
                    "Inheritance": True,
                    "Perms": {"Deny": None, "Grant": None},
                    "Value": reg_vars.vdata,
                    "Key": reg_vars.name,
                    "Owner": None,
                    "Entry": reg_vars.vname,
                }
            }
        },
        "name": reg_vars.name,
        "result": True,
    }
    assert (
        reg.present(name=reg_vars.name, vname=reg_vars.vname, vdata=reg_vars.vdata)
        == expected
    )
    permissions = win_dacl.get_permissions(obj_name=reg_vars.name, obj_type="registry")
    assert permissions["Not Inherited"] == {}


def test_present_set_owner(reg_vars, clean):
    """
    Test reg.present
    """
    reg.present(
        name=reg_vars.name,
        vname=reg_vars.vname,
        vdata=reg_vars.vdata,
        win_owner=reg_vars.current_user,
    )
    assert (
        win_dacl.get_owner(obj_name=reg_vars.name, obj_type="registry")
        == reg_vars.current_user
    )


def test_present_perms_no_inherit(reg_vars, clean):
    reg.present(
        name=reg_vars.name,
        vname=reg_vars.vname,
        vdata=reg_vars.vdata,
        win_inheritance=False,
    )
    assert not win_dacl.get_inheritance(obj_name=reg_vars.name, obj_type="registry")
    permissions = win_dacl.get_permissions(obj_name=reg_vars.name, obj_type="registry")
    assert permissions["Inherited"] == {}


def test_present_perms(reg_vars, clean):
    reg.present(
        name=reg_vars.name,
        vname=reg_vars.vname,
        vdata=reg_vars.vdata,
        win_perms={"Backup Operators": {"perms": "full_control"}},
        win_deny_perms={"Guest": {"perms": "full_control"}},
    )
    permissions = win_dacl.get_permissions(obj_name=reg_vars.name, obj_type="registry")
    expected = {
        "deny": {"permissions": "Full Control", "applies to": "This key and subkeys"}
    }
    assert permissions["Not Inherited"].get("Guest") == expected
    expected = {
        "grant": {"permissions": "Full Control", "applies to": "This key and subkeys"}
    }
    assert permissions["Not Inherited"].get("Backup Operators") == expected


def test_present_perms_reset(reg_vars, clean):
    reg.present(
        name=reg_vars.name,
        vname=reg_vars.vname,
        vdata=reg_vars.vdata,
        win_perms={"Everyone": {"perms": "full_control"}},
        win_deny_perms={"Guest": {"perms": "full_control"}},
        win_perms_reset=True,
    )
    permissions = win_dacl.get_permissions(obj_name=reg_vars.name, obj_type="registry")
    expected = {
        "Guest": {
            "deny": {
                "permissions": "Full Control",
                "applies to": "This key and subkeys",
            }
        },
        "Everyone": {
            "grant": {
                "permissions": "Full Control",
                "applies to": "This key and subkeys",
            }
        },
    }
    assert permissions["Not Inherited"] == expected


def test_present_perms_reset_no_inherit(reg_vars, clean):
    reg.present(
        name=reg_vars.name,
        vname=reg_vars.vname,
        vdata=reg_vars.vdata,
        win_perms={"Everyone": {"perms": "full_control"}},
        win_deny_perms={"Guest": {"perms": "full_control"}},
        win_perms_reset=True,
        win_inheritance=False,
    )
    permissions = win_dacl.get_permissions(obj_name=reg_vars.name, obj_type="registry")
    expected = {
        "Guest": {
            "deny": {
                "permissions": "Full Control",
                "applies to": "This key and subkeys",
            }
        },
        "Everyone": {
            "grant": {
                "permissions": "Full Control",
                "applies to": "This key and subkeys",
            }
        },
    }
    assert permissions["Not Inherited"] == expected
    assert not win_dacl.get_inheritance(obj_name=reg_vars.name, obj_type="registry")
    assert permissions["Inherited"] == {}


def test_present_string_dword(reg_vars, clean):
    """
    Test to set a registry entry.
    """
    vname = "dword_data"
    vdata = "00000001"
    vtype = "REG_DWORD"
    expected_vdata = 1
    expected = {
        "comment": "Added {} to {}".format(vname, reg_vars.name),
        "changes": {
            "reg": {
                "Added": {
                    "Inheritance": True,
                    "Perms": {"Deny": None, "Grant": None},
                    "Value": expected_vdata,
                    "Key": reg_vars.name,
                    "Owner": None,
                    "Entry": vname,
                }
            }
        },
        "name": reg_vars.name,
        "result": True,
    }
    assert (
        reg.present(name=reg_vars.name, vname=vname, vdata=vdata, vtype=vtype)
        == expected
    )


def test_present_string_dword_existing(reg_vars, clean):
    """
    Test to set a registry entry.
    """
    vname = "dword_data"
    vdata = "0000001"
    vtype = "REG_DWORD"
    # Set it first
    reg_util.set_value(
        hive=reg_vars.hive, key=reg_vars.key, vname=vname, vdata=vdata, vtype=vtype
    )
    expected = {
        "comment": "{} in {} is already present".format(vname, reg_vars.name),
        "changes": {},
        "name": reg_vars.name,
        "result": True,
    }
    assert (
        reg.present(name=reg_vars.name, vname=vname, vdata=vdata, vtype=vtype)
        == expected
    )


def test_present_test_true(reg_vars, clean):
    expected = {
        "comment": "",
        "changes": {
            "reg": {
                "Will add": {
                    "Inheritance": True,
                    "Perms": {"Deny": None, "Grant": None},
                    "Value": reg_vars.vdata,
                    "Key": reg_vars.name,
                    "Owner": None,
                    "Entry": "version",
                }
            }
        },
        "name": reg_vars.name,
        "result": None,
    }
    with patch.dict(reg.__opts__, {"test": True}):
        ret = reg.present(reg_vars.name, vname=reg_vars.vname, vdata=reg_vars.vdata)
    assert ret == expected


def test_present_existing(reg_vars, reset):
    expected = {
        "comment": "{} in {} is already present".format(reg_vars.vname, reg_vars.name),
        "changes": {},
        "name": reg_vars.name,
        "result": True,
    }
    assert (
        reg.present(name=reg_vars.name, vname=reg_vars.vname, vdata=reg_vars.vdata)
        == expected
    )


def test_present_existing_key_only(reg_vars, clean):
    """
    Test setting only a key with no value name
    """
    # Create the reg key for testing
    reg_util.set_value(hive=reg_vars.hive, key=reg_vars.key)

    expected = {
        "comment": "(Default) in {} is already present".format(reg_vars.name),
        "changes": {},
        "name": reg_vars.name,
        "result": True,
    }
    assert reg.present(reg_vars.name) == expected


def test_present_existing_test_true(reg_vars, reset):
    expected = {
        "comment": "{} in {} is already present".format(reg_vars.vname, reg_vars.name),
        "changes": {},
        "name": reg_vars.name,
        "result": True,
    }
    with patch.dict(reg.__opts__, {"test": True}):
        ret = reg.present(
            name=reg_vars.name, vname=reg_vars.vname, vdata=reg_vars.vdata
        )
    assert ret == expected


def test_absent(reg_vars, reset):
    """
    Test to remove a registry entry.
    """
    expected = {
        "comment": "Removed {} from {}".format(reg_vars.key, reg_vars.hive),
        "changes": {
            "reg": {"Removed": {"Entry": reg_vars.vname, "Key": reg_vars.name}}
        },
        "name": reg_vars.name,
        "result": True,
    }
    assert reg.absent(reg_vars.name, reg_vars.vname) == expected


def test_absent_test_true(reg_vars, reset):
    expected = {
        "comment": "",
        "changes": {
            "reg": {"Will remove": {"Entry": reg_vars.vname, "Key": reg_vars.name}}
        },
        "name": reg_vars.name,
        "result": None,
    }
    with patch.dict(reg.__opts__, {"test": True}):
        ret = reg.absent(reg_vars.name, reg_vars.vname)
    assert ret == expected


def test_absent_already_absent(reg_vars, clean):
    """
    Test to remove a registry entry.
    """
    expected = {
        "comment": "{} is already absent".format(reg_vars.name),
        "changes": {},
        "name": reg_vars.name,
        "result": True,
    }
    assert reg.absent(reg_vars.name, reg_vars.vname) == expected


def test_absent_already_absent_test_true(reg_vars, clean):
    """
    Test to remove a registry entry.
    """
    expected = {
        "comment": "{} is already absent".format(reg_vars.name),
        "changes": {},
        "name": reg_vars.name,
        "result": True,
    }
    with patch.dict(reg.__opts__, {"test": True}):
        ret = reg.absent(reg_vars.name, reg_vars.vname)
    assert ret == expected



"""
unit tests for the git_pillar runner
"""


import errno
import logging
import tempfile

import salt.runners.git_pillar as git_pillar
import salt.utils.files
import salt.utils.gitfs
from tests.support.gitfs import _OPTS
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import patch
from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import TestCase

log = logging.getLogger(__name__)


class GitPillarTest(TestCase, LoaderModuleMockMixin):
    """
    Validate the git_pillar runner
    """

    @classmethod
    def setUpClass(cls):
        cls.tmp_cachedir = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)

    @classmethod
    def tearDownClass(cls):
        try:
            salt.utils.files.rm_rf(cls.tmp_cachedir)
        except OSError as exc:
            if exc.errno == errno.EACCES:
                log.error("Access error removing file %s", cls.tmp_cachedir)
            elif exc.errno != errno.EEXIST:
                raise

    def setup_loader_modules(self):
        opts = _OPTS.copy()
        opts["cachedir"] = self.tmp_cachedir
        opts["verified_git_pillar_provider"] = "gitfoo"
        opts["ext_pillar"] = [
            {
                "git": [
                    "master https://someurl/some",
                    {"dev https://otherurl/other": [{"name": "somename"}]},
                ]
            }
        ]
        return {git_pillar: {"__opts__": opts}}

    def test_update(self):
        """
        test git_pillar.update
        """

        class MockGitProvider(
            salt.utils.gitfs.GitProvider
        ):  # pylint: disable=abstract-method
            def init_remote(self):
                new = False
                self.repo = True
                return new

            def fetch(self):
                return True

            def clear_lock(self, lock_type="update"):
                pass  # return success, failed

        git_providers = {"gitfoo": MockGitProvider}

        repo1 = {"master https://someurl/some": True}
        repo2 = {"dev https://otherurl/other": True}
        all_repos = {
            "master https://someurl/some": True,
            "dev https://otherurl/other": True,
        }
        with patch.object(salt.utils.gitfs, "GIT_PROVIDERS", git_providers):
            self.assertEqual(git_pillar.update(), all_repos)
            self.assertEqual(git_pillar.update(branch="master"), repo1)
            self.assertEqual(git_pillar.update(branch="dev"), repo2)
            self.assertEqual(git_pillar.update(repo="somename"), repo2)
